import copy
import json
import logging
import math
import os
import random
import time
from dataclasses import asdict
from typing import Literal, Generator, Iterable

from agere.commander import Callback
from ai_care import AICare, AICareContext
from semantic_kernel.connectors.ai.open_ai import OpenAITextEmbedding

from .ai_care_sensors import time_now
from ..gptui_kernel.manager import ManagerInterface
from ..models.blinker_wrapper import async_wrapper_with_loop, async_wrapper_without_loop
from ..models.context import BeadOpenaiContext, OpenaiContext
from ..models.jobs import GroupTalkManager
from ..models.openai_chat import OpenaiChatInterface, OpenAIGroupTalk
from ..models.openai_chat_inner_service import chat_service_for_inner
from ..models.signals import (
    notification_signal,
    response_to_user_message_stream_signal,
    chat_context_extend_signal,
    response_to_user_message_sentence_stream_signal,
)
from ..models.utils.openai_settings_from_dot_env import openai_settings_from_dot_env
from ..models.utils.openai_api import openai_api_client
from ..utils.my_text import MyText as Text
from ..views.animation import AnimationRequest
from ..views.screens import InputDialog


gptui_logger = logging.getLogger("gptui_logger")


class OpenaiChatManage:
    """
    Schema for conversations:
    {
        conversation_id: {
            "tab_name": tab_name,
            "file_id": file_id,
            "openai_context": BeadOpenaiContext,
            "max_sending_tokens_ratio": the proportion or the maximum number of tokens sent to the total tokens window,
        }
    }
    Schema for group talk conversations:
    {
        conversation_id: {
            "tab_name": tab_name,
            "file_id": file_id,
            "group_talk_manager": GroupTalkManager,
            "max_sending_tokens_ratio": the proportion or the maximum number of tokens sent to the total tokens window,
        }
    }
    """

    def __init__(
        self,
        app,
        manager: ManagerInterface,
        openai_chat: OpenaiChatInterface,
        workpath: str,
        conversations_recover: bool = False,
        vector_memory: bool = True,
    ) -> None:
        assert manager.dot_env_config_path is not None
        self.openai_api_key, self.openai_org_id = openai_settings_from_dot_env(manager.dot_env_config_path)
        # check if workpath is a valid directory
        if not os.path.isdir(workpath):
            if os.path.exists(workpath):
                raise ValueError("'workpath' is not a directory.")
            else:
                os.makedirs(workpath)
        self.app = app
        self.openai_chat = openai_chat
        self.manager = manager
        self.workpath = workpath or os.path.join(app.workpath, "conversations")
        self.conversation_dict = {}
        self.conversation_count = 0
        self.conversation_active = 0
        self.group_talk_conversation_active = 0
        self.conversation_id_set = set()
        if conversations_recover:
            try:
                file_path = os.path.join(workpath, '_conversations_cache.json')
                with open(file_path, "r") as read_file:
                    json_str = read_file.read()
                    conversation_cache = json.loads(json_str)
                    conversation_plugins_dict = conversation_cache["conversation_plugins_dict"]
                    old_conversation_dict = conversation_cache["conversation_dict"]
                    new_conversation_dict = {}
                    for key, value in old_conversation_dict.items():
                        # rebuild OpenaiContext
                        openai_context_build = value["openai_context"]
                        # retrieve plugins list
                        openai_context_build["plugins"] = plugins_from_name(
                            manager=manager,
                            plugin_path=app.config["PLUGIN_PATH"],
                            plugins_name_list=conversation_plugins_dict[key],
                        )
                        value["openai_context"] = BeadOpenaiContext(**openai_context_build)
                        # convert id to int
                        new_conversation_dict[int(key)] = value
                        self.conversation_id_set.add(int(key))
                    self.conversation_dict = new_conversation_dict
                    self.conversation_active = int(conversation_cache["conversation_active"])
            except Exception as e:
                text = Text(f"{type(e).__name__}    " + "Read conversation cache failed, opened a new conversation.", "red")
                app.post_message(AnimationRequest(
                    ani_id="read conversation cache failed",
                    action="start",
                    ani_type="static",
                    keep_time=5,
                    priority=0,
                    ani_end_display=text,
                    others=text,
                ))
                gptui_logger.warning(f"Read conversation cache failed, opened a new conversation: {e}")
                id = self.open_conversation_with_mode()
                self.conversation_active = id
        else:
            id = self.open_conversation_with_mode()
            self.conversation_active = id

        if vector_memory is True:
            self.init_volatile_memory()
        
        self.openai_group_talk = OpenAIGroupTalk(manager=manager)
        self.group_talk_conversation_dict = {}
        self.openai_client = openai_api_client(manager.dot_env_config_path)
        self.ai_care = AICare()
        self._set_ai_care(self.ai_care)
        self.accept_ai_care: bool = True
        self.ai_care_depth_default = app.config["tui_config"]["ai_care_depth"]
        self.ai_care_depth: int = self.ai_care_depth_default

    def reset_ai_care_depth(self):
        if self.ai_care_depth_default <= 1:
            self.ai_care_depth = self.ai_care_depth_default
        else:
            self.ai_care_depth = random.randint(1, self.ai_care_depth_default)

    def _set_ai_care(self, ai_care: AICare):
        ai_care.set_config(key="delay", value=self.app.config["tui_config"]["ai_care_delay"])
        ai_care.set_guide("You are a very considerate person who cares about others and is willing to inntiate conversations")
        ai_care.register_to_llm_method(self.ai_care_to_openai)
        ai_care.register_to_user_method(self.ai_care_to_user)
        self._register_sensors(ai_care)

    def _register_sensors(self, ai_care: AICare):
        ai_care.register_sensor(
            name="time_now",
            function=time_now,
            annotation="Get the current date and time in the local time zone"
        )

    def init_volatile_memory(self):
        kernel = self.manager.services.sk_kernel
        kernel.add_text_embedding_generation_service("ada", OpenAITextEmbedding("text-embedding-ada-002", self.openai_api_key, self.openai_org_id or ""))
        kernel.register_memory_store(memory_store=self.app.qdrant_vector)

    def bead_insert(self, conversation_id: int | None = None) -> BeadOpenaiContext:
        "Insert the bead into the chat context."
        if conversation_id is None:
            conversation_id = self.conversation_active
        conversation = self.conversation_dict[conversation_id]
        openai_context = conversation["openai_context"]
        openai_context.insert_bead()
        return openai_context

    def auto_bead_insert(self, conversation_id: int | None = None) -> tuple[BeadOpenaiContext, bool]:
        "Auto insert the bead into the chat context."
        if conversation_id is None:
            conversation_id = self.conversation_active
        conversation = self.conversation_dict[conversation_id]
        openai_context = conversation["openai_context"]
        do_insert = openai_context.auto_insert_bead()
        
        if do_insert:
            app = self.app
            app.main_screen.query_one("#status_region").update(Text("Bead inserted.","green"))
            #refresh dashboard
            model = conversation["openai_context"].parameters["model"]
            app.dash_board.dash_board_display(tokens_num_window=app.get_tokens_window(model))
            return openai_context, True
        return openai_context, False
    
    def bead_init(self, bead_id: int) -> dict:
        """
        Generate a initial template bead.
        """
        bead = {
            "role": "system",
            "content": (
                "Memo:\nYour memory is limited. When encountering important information, you should use memo to record it.\n"
                f"CONVERSATION ID: {bead_id}"
            )
        }
        return bead
    
    def open_conversation(
        self,
        id: int | str | None = None,
        openai_params: dict | None = None,
        max_sending_tokens_ratio: float | None = None
    ) -> int:
        """
        Open a empty conversation, and return the conversation's id.
        Conversation active id should be handle manually
        """
        self.conversation_count = id or time.time() * 1000
        self.conversation_count = int(self.conversation_count)
        while self.conversation_count in self.conversation_id_set or self.conversation_count == 0:
            assert True, "The conversation_id is duplicated."
            self.conversation_count += 1
        bead_init = self.bead_init(self.conversation_count)
        openai_context = BeadOpenaiContext(bead=[bead_init])
        openai_context.id = self.conversation_count
        openai_context.chat_context = []
        openai_context.plugins = plugins_from_name(manager=self.manager, plugin_path=self.app.config["PLUGIN_PATH"], plugins_name_list=self.app.config["default_plugins_used"])
        
        if openai_params is not None:
            openai_context.parameters = openai_params.copy()
        else:
            openai_context.parameters = self.app.config["default_openai_parameters"].copy()
        openai_context.chat_context_saver = "outer"
        openai_context.chat_context_saver_for_sending = "outer"
        
        if max_sending_tokens_ratio is None:
            max_sending_tokens_ratio = self.app.config["default_conversation_parameters"]["max_sending_tokens_ratio"]
        # check max_sending_tokens_num validity
        if not isinstance(max_sending_tokens_ratio, float):
            raise TypeError("'max_sending_tokens_ratio' have to be a flat number.")
        if not 0.0 < max_sending_tokens_ratio < 1.0:
            raise ValueError("'max_sending_tokens_ratio' have to be in range from 0.0 to 1.0.")

        # set max_sending_tokens_num
        tokens_window = self.app.config["openai_model_info"][openai_context.parameters["model"]]["tokens_window"]
        openai_context.max_sending_tokens_num = math.floor(max_sending_tokens_ratio * tokens_window)
        # set max_tokens
        if openai_context.parameters.get("max_tokens") is None:
            openai_context.parameters["max_tokens"] = tokens_window - openai_context.max_sending_tokens_num
        
        self.conversation_dict[self.conversation_count] = {
            "tab_name": "New",
            "file_id": None,
            "openai_context": openai_context,
            "max_sending_tokens_ratio": max_sending_tokens_ratio,
        }
        
        # insert initial bead
        self.bead_insert(self.conversation_count)
        self.conversation_id_set.add(self.conversation_count)
        return self.conversation_count
    
    def open_conversation_with_mode(
        self,
        id: int | str | None = None,
        mode: Literal["Normal"] | None = None,
        openai_params: dict | None = None,
        max_sending_tokens_ratio: float | None = None
        ) -> int:
        """
        open a conversation with mode:
            "Normal"
        """
        if mode is None:
            mode = "Normal"
        conversation_id = self.open_conversation(id=id, openai_params=openai_params, max_sending_tokens_ratio=max_sending_tokens_ratio)
        if mode == "Normal":
            pass
        return conversation_id

    def delete_conversation(self, conversation_id: int = 0) -> None:
        "delete a conversation from conversation dict"
        if conversation_id == 0:
            conversation_id = self.conversation_active
        del self.conversation_dict[conversation_id]
    
    async def write_conversation(self, conversation_id: int) -> bool | Exception:
        "write a conversation to file"
        file_id = self.conversation_dict[conversation_id]["file_id"] 
        if not file_id:
            self.get_file_id_and_save(conversation_id)
            return False
        
        conversation = copy.deepcopy(self.conversation_dict[conversation_id])

        # Do not save plugins information
        # Clear the plugin list to avoid errors, because when the plugin list contains a Manager object, 'asdict' will raise an error.
        conversation["openai_context"].plugins = []
        # Replace openai_context object with dict version for serialization
        conversation["openai_context"] = asdict(conversation["openai_context"])
        
        with open(os.path.join(self.workpath, str(file_id) + '.json'), "w") as write_file:
            try:
                write_file.write(json.dumps({conversation_id: conversation}, ensure_ascii = False, sort_keys = True, indent = 4, separators = (',',':')))
            except Exception as e:
                self.app.main_screen.query_one("#status_region").update(Text(f"Save conversation failed: {e}", "red"))
                gptui_logger.error(f"Write conversation failed. Error: {e}")
                return e
        self.app.main_screen.query_one("#status_region").update(Text(f"Save conversation context successfully.", "green"))
        self.app.main_screen.query_one("#conversation_tree").conversation_refresh()
        return True
    
    def read_conversation(self, file_path: str) -> tuple[bool, Exception | int | str]:
        "load conversation from file"
        if not file_path.endswith('.json'):
            self.app.main_screen.query_one("#status_region").update(Text("Conversation file is not supported.",'yellow'))
            gptui_logger.error("Conversation file is not supported.")
            return False, ValueError("Conversation file is not supported")
        with open(file_path, "r") as read_file:
            try:
                json_str = read_file.read()
                conversation_info = json.loads(json_str)
                conversation_id = list(conversation_info.keys())[0]
                if int(conversation_id) in self.conversation_dict:
                    # The conversation already exists
                    return False, int(conversation_id)
                conversation = conversation_info[conversation_id]
            except FileNotFoundError as e:
                gptui_logger.error("File not found")
                self.app.main_screen.query_one("#status_region").update(Text("File not found",'yellow'))
                return False, e
            except IsADirectoryError as e:
                gptui_logger.error("Specified path is a directory, not a file")
                self.app.main_screen.query_one("#status_region").update(Text("Specified path is a directory, not a file",'yellow'))
                return False, e
            except UnicodeDecodeError as e:
                gptui_logger.error("File is not encoded properly")
                self.app.main_screen.query_one("#status_region").update(Text("File is not encoded properly",'yellow'))
                return False, e
            except IOError as e:
                self.app.main_screen.query_one("#status_region").update(Text(f"An I/O error occurred: {e}",'yellow'))
                gptui_logger.error(f"An I/O error occurred: {e}")
                return False, e
            except Exception as e:
                self.app.main_screen.query_one("#status_region").update(Text('Read conversation failed','red'))
                gptui_logger.error("Read conversation failed. An unexpected error occurred: {e}")
                return False, e
            else:
                openai_context_build = conversation["openai_context"]
                openai_context = BeadOpenaiContext(**openai_context_build)
                openai_parameters = openai_context.parameters
                model = openai_parameters.get("model")
                if model is None:
                    raise ValueError("Field 'model' is not found in conversation.")
                if mode := conversation.get("mode"):
                    id = self.open_conversation_with_mode(id=conversation_id, openai_params=openai_parameters, mode=mode)
                else:
                    id = self.open_conversation(id=conversation_id, openai_params=openai_parameters)
                conversation["openai_context"] = openai_context
                self.conversation_dict[id] = conversation
                self.conversation_active = id
                return True, id

    def get_file_id_and_save(self, conversation_id: int, input_dialog_prompt: str | None = None) -> None:
        "Get file id when write a new context without file id to file"

        # get the existing filenames
        existing_filenames = set()
        for filename in os.listdir(self.workpath):
            if filename.endswith(".json"):
                filename_without_extension = os.path.splitext(filename)[0]
                existing_filenames.add(filename_without_extension)
        
        async def input_handle(input_: tuple[bool, str]) -> None | str:
            status, content = input_
            if status is True:
                if content in existing_filenames:
                    # re-enter
                    self.get_file_id_and_save(conversation_id, input_dialog_prompt="The entered filename already exists, please re-enter.")
                    return
                self.conversation_dict[conversation_id]["file_id"] = content
                status = await self.write_conversation(conversation_id)
                if status is not True:
                    return
                collection = conversation_id
                self.app.qdrant_queue.put(
                    {
                        "action": "collection_save",
                        "content":{
                            "collection_name": str(collection),
                            "event": None,
                        },
                    }
                )
            else:
                return

        tab_name = self.conversation_dict[conversation_id]["tab_name"]
        self.app.push_screen(InputDialog(prompt=input_dialog_prompt or "Enter a name for the conversation:", default_input=tab_name), input_handle)

    def open_group_talk_conversation(
        self,
        id: int | str | None = None,
        max_sending_tokens_ratio: float | None = None,
    ) -> int:
        """
        Open a empty group talk conversation, and return the conversation's id.
        Conversation active id should be handle manually
        """
        self.conversation_count = id or time.time() * 1000
        self.conversation_count = int(self.conversation_count)
        while self.conversation_count in self.conversation_id_set or self.conversation_count == 0:
            assert True, "The conversation_id is duplicated."
            self.conversation_count += 1
        group_talk_manager = GroupTalkManager(manager=self.manager)
        group_talk_manager.group_talk_manager_id = self.conversation_count
        callback = Callback(
            at_job_start=[
                {
                    "function": notification_signal.send,
                    "params": {
                        "args": (self,),
                        "kwargs": {
                            "_async_wrapper": async_wrapper_with_loop,
                            "message": {
                                "content": {
                                    "content": {"status": True, "group_talk_manager": group_talk_manager},
                                    "description": "GroupTalkManager status changed",
                                },
                                "flag": "info",
                            },
                        },
                    },
                },
            ],
            at_terminate=[
                {
                    "function": notification_signal.send,
                    "params": {
                        "args": (self,),
                        "kwargs": {
                            "_async_wrapper": async_wrapper_with_loop,
                            "message": {
                                "content": {
                                    "content": {"status": False, "group_talk_manager": group_talk_manager},
                                    "description": "GroupTalkManager status changed",
                                },
                                "flag": "info",
                            },
                        },
                    },
                },
            ],
            at_job_end=[
                {
                    "function": notification_signal.send,
                    "params": {
                        "args": (self,),
                        "kwargs": {
                            "_async_wrapper": async_wrapper_with_loop,
                            "message": {
                                "content": {
                                    "content": {"status": False, "group_talk_manager": group_talk_manager},
                                    "description": "GroupTalkManager status changed",
                                },
                                "flag": "info",
                            },
                        },
                    },
                },
            ],
        )
        group_talk_manager.add_callback(callback)
        self.group_talk_conversation_dict[self.conversation_count] = {
            "tab_name": "Group-Talk",
            "file_id": None,
            "group_talk_manager": group_talk_manager,
            "max_sending_tokens_ratio": max_sending_tokens_ratio or self.app.config["default_conversation_parameters"]["max_sending_tokens_ratio"],
        }
        self.conversation_id_set.add(self.conversation_count)
        self.group_talk_conversation_active = self.conversation_count
        return self.conversation_count

    def delete_group_talk_conversation(self, group_talk_conversation_id: int = 0) -> None:
        """delete a group talk conversation from group_talk_conversation_dict"""
        if group_talk_conversation_id == 0:
            group_talk_conversation_id = self.group_talk_conversation_active
        group_talk_manager = self.group_talk_conversation_dict[group_talk_conversation_id]["group_talk_manager"]
        group_talk_manager.close_group_talk()
        del self.group_talk_conversation_dict[group_talk_conversation_id]

    def ai_care_to_openai(
        self,
        chat_context: OpenaiContext,
        to_llm_messages: list[AICareContext]
    ) -> Generator[str, None, None]:
        messages_list = [
            {"role": "user", "name": "Aicarey", "content": message["content"]} if message["role"] == "ai_care"
            else {"role": "assistant", "content": message["content"]}
            for message in to_llm_messages
        ]
        openai_response = chat_service_for_inner(
            messages_list=messages_list,
            context=chat_context,
            openai_api_client=self.openai_client,
        )
        def response_gen(response: Iterable):
            for chunk in response:
                content = chunk.choices[0].delta.content
                if content is None:
                    continue
                yield content
        return response_gen(openai_response)

    def ai_care_to_user(self, to_user_message: Generator[str, None, None]) -> None:
        if self.accept_ai_care is False:
            return
        context_id = self.conversation_active
        openai_context = self.conversation_dict[self.conversation_active]["openai_context"]
        char_list = []
        voice_buffer = ""
        first_times = True
        for char in to_user_message:
            response_to_user_message_stream_signal.send(
                self,
                _async_wrapper=async_wrapper_without_loop,
                message={"content": {"content": char.lstrip() if first_times else char, "context_id": context_id}, "flag": "content"},
            )
            char_list.append(char)
            first_times = False
            
            # Send voice signal
            if response_to_user_message_sentence_stream_signal.receivers:
                voice_buffer += char
                if char.startswith((".","!","?",";",":","。","！","？","；","：","\n")):
                    response_to_user_message_sentence_stream_signal.send(
                        self,
                        _async_wrapper=async_wrapper_without_loop,
                        message={"content": voice_buffer, "flag": "content"},
                    )
                    voice_buffer = ""

        response_to_user_message_stream_signal.send(
            self,
            _async_wrapper=async_wrapper_without_loop,
            message={"content": {"content": "", "context_id": context_id}, "flag": "end"},
        )

        chat_context_extend_signal.send(
            self,
            _async_wrapper=async_wrapper_without_loop,
            message={
                "content": {
                    "messages": [{"role": "assistant", "content": ''.join(char_list)}],
                    "context": openai_context,
                },
                "flag": "",
            }
        )
        
        if voice_buffer:
            response_to_user_message_sentence_stream_signal.send(
                self,
                _async_wrapper=async_wrapper_without_loop,
                message={"content": voice_buffer, "flag": "content"},
            )
            voice_buffer = ""

        # Send voice end signal
        if response_to_user_message_sentence_stream_signal.receivers:
            response_to_user_message_sentence_stream_signal.send(
                self,
                _async_wrapper=async_wrapper_without_loop,
                message={"content":"", "flag":"end"}
            )
        
        if self.ai_care_depth > 0:
            self.ai_care.chat_update(openai_context)
            self.ai_care_depth -= 1


def plugins_from_name(manager: ManagerInterface, plugin_path: str, plugins_name_list) -> list[tuple]:
    semantic_plugins_list, native_plugins_list = manager.scan_plugins(plugin_path)
    semantic_plugins_dict = {plugin.plugin_info[1]: plugin.plugin_info for plugin in semantic_plugins_list}
    native_plugins_dict = {plugin.plugin_info[2]: plugin.plugin_info for plugin in native_plugins_list}
    # retrieve plugin_info
    default_plugins_used_ready = [d[k] for d in [semantic_plugins_dict, native_plugins_dict] for k in plugins_name_list if k in d]
    return default_plugins_used_ready



from __future__ import annotations
import logging
import openai
import string
import uuid
from typing import TYPE_CHECKING

from ..models.signals import notification_signal
from ..models.openai_chat import OpenaiContext
from ..utils.my_text import MyText as Text
from ..views.animation import AnimationRequest
from ..views.theme import theme_color as tc


if TYPE_CHECKING:
    from ..views.tui import MainApp


gptui_logger = logging.getLogger("gptui_logger")


class Notification:
    def __init__(self, app: MainApp):
        self.app = app
        self.displayer = app.main_screen.query_one("#status_region")
        notification_signal.connect(self.notification_display)
        self.commander_status = {}

    async def notification_display(self, sender, **kwargs) -> None:
        notification = kwargs["message"]
        content = notification["content"]
        flag = notification["flag"]

        if flag == "info":
            info_content = content["content"]
            description = remove_punctuation(content["description"]).lower()
            if description == "raw":
                self.displayer.update(Text(info_content, tc("green") or "green"))
            elif description == "starting to send the original chat message from the user":
                context = info_content["context"]
                self.app.post_message(
                    AnimationRequest(
                        ani_id=context.id,
                        action="start",
                        ani_type=self.app.config["tui_config"]["waiting_receive_animation"],
                    )
                )
            elif description == "an error occurred in communication with openai initiated by user":
                context = info_content["context"]
                self.app.post_message(AnimationRequest(ani_id=context.id, action="end"))
            elif description == "starting to receive the original response message to the user":
                context = info_content["context"]
                self.app.post_message(AnimationRequest(ani_id=context.id, action="end"))
            elif description == "commander exit":
                commander_status_display = self.app.main_screen.query_one("#commander_status_display")
                commander_status_display.update(Text("\u260a", tc("cyan") or "cyan"))
            elif description == "job status changed":
                await self._handle_job_status_changed(info_content=info_content)
            elif description == "grouptalkmanager status changed":
                tab_id = int(self.app.main_screen.query_one("#chat_tabs").active[3:])
                status = info_content["status"]
                group_talk_manager_id = info_content["group_talk_manager"].group_talk_manager_id
                commander_status_display = self.app.main_screen.query_one("#commander_status_display")
                self.commander_status[group_talk_manager_id] = status
                if tab_id == group_talk_manager_id:
                    if status is True:
                        commander_status_display.update(Text("\u2725", tc("green") or "green"))
                        self.displayer.update(Text("Group talk created.", tc("green") or "green"))
                    else:
                        commander_status_display.update(Text("\u2668", tc("red") or "red"))
                        self.displayer.update(Text("Group talk closed.", tc("green") or "green"))

        elif flag == "warning":
            ani_id = uuid.uuid4()
            self.app.post_message(
                AnimationRequest(
                    ani_id=ani_id,
                    action="start",
                    ani_type="static",
                    keep_time=3,
                    ani_end_display=self.app.status_region_default,
                    others=Text(f"{content}", tc("yellow") or "yellow"),
                )
            )

        elif flag == "error":
            ani_id = uuid.uuid4()
            self.app.post_message(
                AnimationRequest(
                    ani_id=ani_id,
                    action="start",
                    ani_type="static",
                    keep_time=3,
                    ani_end_display=self.app.status_region_default,
                    others=Text(f"{content}", tc("red") or "red"),
                )
            )
        
        elif flag == "openai_error":
            error = content["error"]
            context = content["context"]
            self.openai_error_display(error=error, context=context)

    async def _handle_job_status_changed(self, info_content: dict):
        tab_id = int(self.app.main_screen.query_one("#chat_tabs").active[3:])
        status = info_content["status"]
        context_id = info_content["context"].id
        commander_status_display = self.app.main_screen.query_one("#commander_status_display")
        self.commander_status[context_id] = status
        
        if tab_id != context_id:
            return
        
        if status is True:
            commander_status_display.update(Text("\u260d", tc("red") or "red"))
            return
        
        self.displayer.update(self.app.status_region_default)
        commander_status_display.update(Text("\u260c", tc("yellow") or "yellow"))
        await self.app.chat_context.chat_context_vectorize()
        commander_status_display.update(Text("\u260c", tc("green") or "green"))
        
        conversation = self.app.openai.conversation_dict.get(context_id)
        if conversation is None:
            return
        
        # AICare
        if self.app.main_screen.query_one("#ai_care_switch").value is True:
            self.app.openai.accept_ai_care = True
            if self.app.openai.ai_care_depth > 0:
                self.app.openai.ai_care.chat_update(conversation["openai_context"])
                self.app.openai.ai_care_depth -= 1
        
        # Tab rename
        context = conversation["openai_context"]
        self.app.conversation_tab_rename(context)

    def openai_error_display(self, error: Exception, context: OpenaiContext) -> None:
        app = self.app
        keep_time = 1
        ani_id = context.id
        if isinstance(error, openai.APIStatusError):
            text = Text(f"OpenAI APIStatusError: {error}", tc("red") or "red")
            app.post_message(AnimationRequest(ani_id = ani_id, action = "start", ani_type="static", keep_time=keep_time, priority=0, ani_end_display=text, others=text))
            gptui_logger.error(f"OpenAI APIStatusError: {error}")
        elif isinstance(error, openai.APITimeoutError):
            text = Text(f"OpenAI APITimeoutError: {error}", tc("red") or "red")
            app.post_message(AnimationRequest(ani_id = ani_id, action = "start", ani_type="static", keep_time=keep_time, priority=0, ani_end_display=text, others=text))
            gptui_logger.error(f"OpenAI APITimeoutError: {error}")
        elif isinstance(error, openai.APIConnectionError):
            text = Text(f"OpenAI APIConnectionError: {error}", tc("red") or "red")
            app.post_message(AnimationRequest(ani_id = ani_id, action = "start", ani_type="static", keep_time=keep_time, priority=0, ani_end_display=text, others=text))
            gptui_logger.error(f"OpenAI APIConnectionError: {error}")
        else:
            text = Text(f"Unknown error occurred: {error}", tc("red") or "red")
            app.post_message(AnimationRequest(ani_id = ani_id, action = "start", ani_type="static", keep_time=keep_time, priority=0, ani_end_display=text, others=text))
            gptui_logger.error(f"Unknown error occurred: {error}")


def remove_punctuation(s: str):
    return s.translate(str.maketrans('', '', string.punctuation))



import logging

import aiofiles

from ..models.doc import Doc
from ..models.skills import UploadFile
from ..utils.my_text import MyText as Text
from ..views.theme import theme_color as tc


gptui_logger = logging.getLogger("gptui_logger")


class TubeFiles:
    def __init__(self, displayer):
        self.displayer = displayer

    async def insert_files(self, *docs: Doc, input: str) -> str:
        if not docs:
            return input
        upload_file = UploadFile()
        return await upload_file.import_file_to_context(*docs, input=input)

    async def read_file_async(self, path: str, encoding="UTF-8") -> None | str:
        """
        Read a file
        Args:
            path -- The path to the file to read
        Returns:
            The contents of the file
        """
        try:
            async with aiofiles.open(path, "r", encoding=encoding) as fp:
                content = await fp.read()
        except FileNotFoundError:
            gptui_logger.error("File or directory not found")
            self.displayer.update(Text("File or directory not found", tc("yellow") or "yellow"))
            return
        except IsADirectoryError:
            gptui_logger.error("Specified path is a directory, not a file")
            self.displayer.update(Text("Specified path is a directory, not a file", tc("yellow") or "yellow"))
            return
        except UnicodeDecodeError:
            gptui_logger.error("File is not encoded properly")
            self.displayer.update(Text("File is not encoded properly", tc("yellow") or "yellow"))
            return
        except IOError as e:
            gptui_logger.error(f"An I/O error occurred: {e}")
            self.displayer.update(Text("An I/O error occurred", tc("yellow") or "yellow"))
            return
        except Exception as e:
            gptui_logger.error(f"Read file failed. An unexpected error occurred: {e}")
            self.displayer.update(Text(f"Read file failed. An unexpected error occurred: {e}", tc("yellow") or "yellow"))
            return
        else:
            return content

    async def write_file_async(self, file_path: str, file_content) -> bool:
        """
        Write a file
        """
        assert file_content is not None, "Content is required and should not be empty"
        assert file_path is not None, "Path is required and should not be empty"
        try:
            async with aiofiles.open(file_path, "w") as fp:
                await fp.write(file_content)
        except FileNotFoundError:
            gptui_logger.error("File or directory not found")
            self.displayer.update(Text("File or directory not found", tc("yellow") or "yellow"))
            return False
        except IsADirectoryError:
            gptui_logger.error("Specified path is a directory, not a file")
            self.displayer.update(Text("Specified path is a directory, not a file", tc("yellow") or "yellow"))
            return False
        except IOError as e:
            gptui_logger.error(f"An I/O error occurred: {e}")
            self.displayer.update(Text("An I/O error occurred", tc("yellow") or "yellow"))
            return False
        except Exception as e:
            gptui_logger.error(f"Read file failed. An unexpected error occurred: {e}")
            self.displayer.update(Text(f"Read file failed. An unexpected error occurred: {e}", tc("yellow") or "yellow"))
            return False
        else:
            return True



import logging

from semantic_kernel.orchestration.sk_context import SKContext
from semantic_kernel.skill_definition import sk_function, sk_function_context_parameter

from gptui.gptui_kernel.manager import auto_init_params
from gptui.models.role import Role


gptui_logger = logging.getLogger("gptui_logger")


class GroupTalk:
    def __init__(self, manager):
        self.manager = manager
    
    @auto_init_params("0")
    @classmethod
    def get_init_params(cls, manager) -> tuple:
        return (manager,)

    @sk_function(
        description="Initialize a group talk",
        name="create_group_talk",
    )
    async def create_group_talk(self) -> str:
        group_talk_id = await self.manager.client.open_group_talk()
        return (
            "Group talk have been created successfully."
            f"The ID of the group talk is {group_talk_id}. This ID is important; "
            "you should remember it and not confuse it with the conversation ID you already have."
            "Next, you should create roles for this group talk. "
            "If the user has not yet specified roles, you should ask the user what kind of roles he would like to create."
        )

    @sk_function(
        description="Close the group talk",
        name="close_group_talk",
        input_description="The ID of the group talk",
    )
    def close_group_talk(self, group_talk_id: str) -> str:
        group_talk_manager = self.manager.client.openai.group_talk_conversation_dict[int(group_talk_id)]["group_talk_manager"]
        group_talk_manager.close_group_talk()
        return "The group talk has been closed."

    @sk_function(
        description="Create a group talk role",
        name="create_group_talk_role",
    )
    @sk_function_context_parameter(
        name="role_name",
        description="The name of the role, must match '^[a-zA-Z0-9_-]{1,64}$', spaces are not allowed in the role_name.",
    )
    @sk_function_context_parameter(
        name="group_talk_id",
        description="The ID of the group talk"
    )
    @sk_function_context_parameter(
        name="role_description",
        description=(
            "The character description in the second person, telling the character his (or her) name, identity, and essential background information, etc.\n"
            "An example:\n"
            "Your name is Thomas, you are a physicist, you always think and judge problems rationally. "
            "You should participate in discussions in a neutral and objective manner. "
            "You are not talkative, but you should use your knowledge in physics to help solve problems."
        )
    )
    def create_group_talk_role(self, context: SKContext) -> str:
        role_name = context["role_name"]
        role_description = context["role_description"]
        group_talk_id = int(context["group_talk_id"])
        if not role_name or not role_description:
            return "Both 'role_name' and 'role_description' is necessary."
        try:
            group_talk_conversation = self.manager.client.openai.group_talk_conversation_dict[group_talk_id]
        except KeyError:
            return f"There is no corresponding ID for {group_talk_id}."
        group_talk_manager = group_talk_conversation["group_talk_manager"]
        prompt = f"You are in a multi-person conversation setting. You are {role_name}."
        prompt += role_description
        prompt += (
            "Your task is to engage in conversations, provide opinions, and interact with other participants in a manner consistent with your role's identity setting. "
            "Based on the context of the conversation, cautiously assess whether you truly have something to say. "
            "If there is no need to speak, simply reply with a space. "
            "Before speaking, first ask the host 'Can I speak?'(send the literal phrase 'Can I speak?') and only speak after receiving permission. "
            "If granted the right to speak, directly state what you intend to say without making unrelated remarks such as thanking the host. "
            f"Please engage in the chat while adhering to these guidelines and staying in character as the '{role_name}'."
        )
        openai_context_parent = self.manager.client.openai.conversation_dict[self.manager.client.openai.conversation_active]["openai_context"]
        role = Role(
            name=role_name,
            group_talk_manager=group_talk_manager,
            manager=self.manager,
            openai_context_parent=openai_context_parent,
        )
        role.set_role_prompt(prompt)
        group_talk_manager.create_role(role=role, role_name=role_name)

        return f"The role of {role_name} has been created."

    @sk_function(
        description="Delete a group talk role",
        name="delete_group_talk_role",
    )
    @sk_function_context_parameter(
        name="role_name",
        description="The name of the role, must match '^[a-zA-Z0-9_-]{1,64}$'",
    )
    @sk_function_context_parameter(
        name="group_talk_id",
        description="The ID of the group talk"
    )
    def delete_group_talk_role(self, context: SKContext) -> str:
        role_name = context["role_name"]
        group_talk_id = int(context["group_talk_id"])
        try:
            group_talk_conversation = self.manager.client.openai.group_talk_conversation_dict[group_talk_id]
        except KeyError:
            return f"There is no corresponding ID for {group_talk_id}."
        roles = group_talk_conversation["group_talk_manager"].roles
        if role_name in roles:
            del roles[role_name]
            return "The role '{role_name}' has been deleted."
        else:
            return f"Role name {role_name} dose not exist."
    
    @sk_function(
        description="Add explanatory prompts for the character's features",
        name="add_role_prompt",
    )
    @sk_function_context_parameter(
        name="role_name",
        description="The name of the role, must match '^[a-zA-Z0-9_-]{1,64}$'",
    )
    @sk_function_context_parameter(
        name="role_prompt",
        description="Prompt for indicating the role's identity and tasks"
    )
    @sk_function_context_parameter(
        name="group_talk_id",
        description="The ID of the group talk"
    )
    def add_role_prompt(self, context: SKContext) -> str:
        role_name = context["role_name"]
        role_prompt = context["role_prompt"]
        group_talk_id = int(context["group_talk_id"])
        try:
            group_talk_conversation = self.manager.client.openai.group_talk_conversation_dict[group_talk_id]
        except KeyError:
            return f"There is no corresponding ID for {group_talk_id}."
        roles = group_talk_conversation["group_talk_manager"].roles
        try:
            role = roles[role_name]
        except KeyError:
            return f"Role name {role_name} dose not exist."
        if role.context.bead:
            role.context.bead[-1]["content"] += "\n" + role_prompt
        else:
            role.set_role_prompt(role_prompt)
        return f"Role prompt has been added to {role_name}."
    
    @sk_function(
        description="Reset role prompts for the character's features",
        name="reset_role_prompt",
    )
    @sk_function_context_parameter(
        name="role_name",
        description="The name of the role, must match '^[a-zA-Z0-9_-]{1,64}$'",
    )
    @sk_function_context_parameter(
        name="role_prompt",
        description="Prompt for indicating the role's identity and tasks"
    )
    @sk_function_context_parameter(
        name="group_talk_id",
        description="The ID of the group talk"
    )
    def reset_role_prompt(self, context: SKContext) -> str:
        role_name = context["role_name"]
        role_prompt = context["role_prompt"]
        group_talk_id = int(context["group_talk_id"])
        try:
            group_talk_conversation = self.manager.client.openai.group_talk_conversation_dict[group_talk_id]
        except KeyError:
            return f"There is no corresponding ID for {group_talk_id}."
        roles = group_talk_conversation["group_talk_manager"].roles
        try:
            role = roles[role_name]
        except KeyError:
            return f"Role name {role_name} dose not exist."
        role.set_role_prompt(role_prompt)
        return f"Role prompt has been added to {role_name}."



import logging
import math
import random
import textwrap

from rich.text import TextType

from .wink_wink import Happy, Apple
from .mywidgets import MyChatWindow
from ..utils.safe_iterate import safe_send
from ..utils.my_text import MyText as Text


gptui_logger = logging.getLogger("gptui_logger")


class FunZone(Happy):
    
    def __init__(self, displayer: MyChatWindow):
        self.displayer = displayer

    @property
    def canvas_width(self) -> int:
        return self.displayer.content_size.width
    
    @property
    def canvas_height(self) -> int:
        return self.displayer.content_size.height

    def refresh(self, content: str | Text, apple_width: int, apple_height: int):
        top_margin = (self.canvas_height - apple_height) // 2
        text = Text("\n" * top_margin)
        if isinstance(content, str):
            content = Text(content)
        content_lines = content.split()
        for line in content_lines:
            line.align("center", self.canvas_width)
        content_after_center = Text("\n").join(content_lines)
        text.append_text(content_after_center)
        self.displayer.update(text, scroll_to_end=False)


class JustBeing(Apple):
    def __init__(self):
        self.text = Text(
            """
            I am not your AI assistant.
            I talk when I want to.
            I do what I want to do.
            I am just a being, here and now.
            """,
            "italic",
        )
        
    def frame(self, inp) -> tuple[bool, tuple[float, TextType]]:
        return False, (0, self.text)

    @property
    def canvas_width(self) -> int:
        return 70
    
    @property
    def canvas_height(self) -> int:
        return 3


class BombBoom(Apple):
    def __init__(self):
        self.frame_gen = self._frame()
        self.frame_gen.send(None)
        self.nums = [
            Text(textwrap.dedent(
                    """
                         999999999     
                       99:::::::::99   
                     99:::::::::::::99 
                    9::::::99999::::::9
                    9:::::9     9:::::9
                    9:::::9     9:::::9
                     9:::::99999::::::9
                      99::::::::::::::9
                        99999::::::::9 
                             9::::::9  
                            9::::::9   
                           9::::::9    
                          9::::::9     
                         9::::::9      
                        9::::::9       
                       99999999        
                    """
                ), "green"
            ),
            Text(textwrap.dedent(
                    """
                         888888888     
                       88:::::::::88   
                     88:::::::::::::88 
                    8::::::88888::::::8
                    8:::::8     8:::::8
                    8:::::8     8:::::8
                     8:::::88888:::::8 
                      8:::::::::::::8  
                     8:::::88888:::::8 
                    8:::::8     8:::::8
                    8:::::8     8:::::8
                    8:::::8     8:::::8
                    8::::::88888::::::8
                     88:::::::::::::88 
                       88:::::::::88   
                         888888888     
                    """
                ), "green"
            ),
            Text(textwrap.dedent(
                    """
                    77777777777777777777
                    7::::::::::::::::::7
                    7::::::::::::::::::7
                    777777777777:::::::7
                               7::::::7 
                              7::::::7  
                             7::::::7   
                            7::::::7    
                           7::::::7     
                          7::::::7      
                         7::::::7       
                        7::::::7        
                       7::::::7         
                      7::::::7          
                     7::::::7           
                    77777777            
                    """
                ), "green"
            ),
            Text(textwrap.dedent(
                    """
                            66666666   
                           6::::::6    
                          6::::::6     
                         6::::::6      
                        6::::::6       
                       6::::::6        
                      6::::::6         
                     6::::::::66666    
                    6::::::::::::::66  
                    6::::::66666:::::6 
                    6:::::6     6:::::6
                    6:::::6     6:::::6
                    6::::::66666::::::6
                     66:::::::::::::66 
                       66:::::::::66   
                         666666666     
                    """
                ), "yellow"
            ),
            Text(textwrap.dedent(
                    """
                    555555555555555555 
                    5::::::::::::::::5 
                    5::::::::::::::::5 
                    5:::::555555555555 
                    5:::::5            
                    5:::::5            
                    5:::::5555555555   
                    5:::::::::::::::5  
                    555555555555:::::5 
                                5:::::5
                                5:::::5
                    5555555     5:::::5
                    5::::::55555::::::5
                     55:::::::::::::55 
                       55:::::::::55   
                         555555555     
                    """
                ), "yellow"
            ),
            Text(textwrap.dedent(
                    """
                           444444444  
                          4::::::::4  
                         4:::::::::4  
                        4::::44::::4  
                       4::::4 4::::4  
                      4::::4  4::::4  
                     4::::4   4::::4  
                    4::::444444::::444
                    4::::::::::::::::4
                    4444444444:::::444
                              4::::4  
                              4::::4  
                              4::::4  
                            44::::::44
                            4::::::::4
                            4444444444
                    """
                ), "yellow"
            ),
            Text(textwrap.dedent(
                    """
                     333333333333333   
                    3:::::::::::::::33 
                    3::::::33333::::::3
                    3333333     3:::::3
                                3:::::3
                                3:::::3
                        33333333:::::3 
                        3:::::::::::3  
                        33333333:::::3 
                                3:::::3
                                3:::::3
                                3:::::3
                    3333333     3:::::3
                    3::::::33333::::::3
                    3:::::::::::::::33 
                     333333333333333   
                    """
                ), "red"
            ),
            Text(textwrap.dedent(
                    """
                     222222222222222    
                    2:::::::::::::::22  
                    2::::::222222:::::2 
                    2222222     2:::::2 
                                2:::::2 
                                2:::::2 
                             2222::::2  
                        22222::::::22   
                      22::::::::222     
                     2:::::22222        
                    2:::::2             
                    2:::::2             
                    2:::::2       222222
                    2::::::2222222:::::2
                    2::::::::::::::::::2
                    22222222222222222222
                    """
                ), "red"
            ),
            Text(textwrap.dedent(
                    """
                      1111111   
                     1::::::1   
                    1:::::::1   
                    111:::::1   
                       1::::1   
                       1::::1   
                       1::::1   
                       1::::l   
                       1::::l   
                       1::::l   
                       1::::l   
                       1::::l   
                    111::::::111
                    1::::::::::1
                    1::::::::::1
                    111111111111
                    """
                ), "red"
            ),
        ]
        self.boom = textwrap.dedent(
            """
            XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX 
            X X                           X X
            X   X                       X   X
            X     X                   X     X
            X       X               X       X
            X         X           X         X
            X           X       X           X
            X             x   X             X
            X               X               X
            X             X    X            X
            X           X        X          X
            X         X            X        X
            X       X                X      X
            X     X                    X    X
            X   X                        X  X
            X X                            XX
            XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
            """
        )


    def _frame(self):
        answer = random.randint(1, 3)
        inp = yield
        if inp != 0:
            if inp == answer:
                return (0.1, Text(""))
            else:
                for i in range(20):
                    if i % 2 == 0:
                        inp = yield (0.1, Text(self.boom))
                    else:
                        inp = yield (0.1, Text(self.boom, "reverse"))
                    if inp == answer:
                        return (0.1, Text(""))
        for i in self.nums:
            inp = yield (0.5, i)
            if inp != 0:
                if inp == answer:
                    return (0.1, Text(""))
                else:
                    break
        for i in range(20):
            if i % 2 == 0:
                inp = yield (0.1, Text(self.boom))
            else:
                inp = yield (0.1, Text(self.boom, "reverse"))

    def frame(self, inp) -> tuple[bool, tuple[float, TextType]]:
        if inp == "green":
            user_inp = 1
        elif inp == "yellow":
            user_inp = 2
        elif inp == "red":
            user_inp = 3
        else:
            user_inp = 0
        status, result = safe_send(self.frame_gen, user_inp)
        if status == "OK":
            return True, result
        else:
            return False, result

    @property
    def canvas_width(self) -> int:
        return 70
    
    @property
    def canvas_height(self) -> int:
        return 16


class RotatingCube(Apple):

    def __init__(self, happy_width: int = 40, happy_height: int = 80):
        # The width and height of the apple canvas
        self.height = happy_height
        self.width = happy_width
        # A buffer to store the Z coordinate (depth) for each pixel on the canvas
        self.z_buffer = [0] * (self.width * self.height)
        # A buffer to store the character to be drawn for each pixel on the canvas
        self.buffer = [' '] * (self.width * self.height)
        # Horizontal offset for translating the cube's position along the X-axis
        self.horizontal_offset = 0
        # The speed by which the cube increments its position in each iteration
        # The density of points on the surface
        self.increment_speed = 0.6
        # The width of the cube in the 3D coordinate system
        self.cube_width = min(self.width, self.height) // 2
        # A constant to control the scaling factor for projecting 3D points onto the 2D screen
        self.K1 = self.cube_width * 2
        # Distance of the camera from the origin of the 3D coordinate system
        self.distance_from_cam = self.cube_width * 5

        self.frame_gen = self._frame()
        self.frame_gen.send(None)

    @property
    def canvas_width(self) -> int:
        return self.width
    
    @property
    def canvas_height(self) -> int:
        return self.height

    def frame(self, inp) -> tuple[bool, tuple[float, TextType]]:
        status, result = safe_send(self.frame_gen, inp)
        if status == "OK":
            return True, result
        else:
            return False, result
    
    def _frame(self):
        a = random.uniform(0, 2 * math.pi)
        b = random.uniform(0, 2 * math.pi)
        c = random.uniform(0, 2 * math.pi)
        while True:
            yield (0.1, self._draw_cube(a, b, c))
            a += 0.01
            b += 0.02
            c += 0.03

    def _float_range(self, start, stop, step):
        while start < stop:
            yield start
            start += step
    
    def _calculate_x(self, i, j, k, A, B, C):
        return j * math.sin(A) * math.sin(B) * math.cos(C) - k * math.cos(A) * math.sin(B) * math.cos(C) + j * math.cos(A) * math.sin(C) + k * math.sin(A) * math.sin(C) + i * math.cos(B) * math.cos(C)

    def _calculate_y(self, i, j, k, A, B, C):
        return j * math.cos(A) * math.cos(C) + k * math.sin(A) * math.cos(C) - j * math.sin(A) * math.sin(B) * math.sin(C) + k * math.cos(A) * math.sin(B) * math.sin(C) - i * math.cos(B) * math.sin(C)

    def _calculate_z(self, i, j, k, A, B):
        return k * math.cos(A) * math.cos(B) - j * math.sin(A) * math.cos(B) + i * math.sin(B)

    def _calculate_for_surface(self, cube_x, cube_y, cube_z, ch, A, B, C):
        x = self._calculate_x(cube_x, cube_y, cube_z, A, B, C)
        y = self._calculate_y(cube_x, cube_y, cube_z, A, B, C)
        z = self._calculate_z(cube_x, cube_y, cube_z, A, B) + self.distance_from_cam

        # One Over Z
        ooz = 1 / z
        width = self.width
        height = self.height
        K1 = self.K1

        xp = int(width / 2 + self.horizontal_offset + K1 * ooz * x * 2)
        yp = int(height / 2 + K1 * ooz * y)

        idx = xp + yp * width

        if 0 <= idx < width * height:
            if ooz > self.z_buffer[idx]:
                self.z_buffer[idx] = ooz
                self.buffer[idx] = ch

    def _draw_cube(self, A, B, C):

        for i in range(len(self.buffer)):
            self.buffer[i] = ' '
            self.z_buffer[i] = 0
        
        cube_width = self.cube_width
        increment_speed = self.increment_speed

        for cube_x in self._float_range(-cube_width, cube_width, increment_speed):
            for cube_y in self._float_range(-cube_width, cube_width, increment_speed):
                self._calculate_for_surface(cube_x, cube_y, -cube_width, '?', A, B, C)
                self._calculate_for_surface(cube_width, cube_y, cube_x, '*', A, B, C)
                self._calculate_for_surface(-cube_width, cube_y, -cube_x, '=', A, B, C)
                self._calculate_for_surface(-cube_x, cube_y, cube_width, '#', A, B, C)
                self._calculate_for_surface(cube_x, -cube_width, -cube_y, '!', A, B, C)
                self._calculate_for_surface(cube_x, cube_width, cube_y, '+', A, B, C)

        lines = []
        for i in range(self.height):
            line = ''.join(self.buffer[i * self.width : (i + 1) * self.width])
            lines.append(line)
        buffer_str = '\n'.join(lines)
        return buffer_str




""" wait for open-interpreter to be compatible with openai version 1.1.1
import asyncio
import logging

from semantic_kernel.skill_definition import sk_function

from gptui.utils.open_interpreter import MyInterpreter, response_render
from gptui.utils.safe_iterate import safe_next, safe_send


gptui_logger = logging.getLogger("gptui_logger")


class OpenInterpreter:

    def __init__(self):
        self.interpreter = MyInterpreter()
        self.in_chat = False
        self.chat = None
        self.result = None

    @sk_function(
        description=(
            "A code assistant that allows for continuous dialogue in natural language. "
            "It can be invoked continuously multiple times"
            "Describe your needs to it, and it will automatically write and execute code to help you accomplish tasks. "
            "When asked whether to execute the code, respond to this function precisely with 'y' or 'n'. "
            "Before responding with 'y', you should first seek the user's consent."
        ),
        name="open_interpreter",
        input_description="Your needs.",
    )
    async def open_interpreter(self, input_request: str) -> str:
        if not self.in_chat:
            self.chat = self.interpreter.chat(str(input_request))
            status, out = await asyncio.to_thread(safe_next, self.chat)
            if status == "OK":
                self.in_chat = True
            else:
                self.in_chat = False
            result = response_render(out)
            gptui_logger.info(f"Open interpreter response: {result}")
            self.new_chat = False
            return result
        else:
            assert self.chat is not None
            status, out = await asyncio.to_thread(safe_send, self.chat, str(input_request))
            if status == "OK":
                self.in_chat = True
            else:
                self.in_chat = False
            result = response_render(out)
            gptui_logger.info(f"Open interpreter response: {result}")
            return result

    @sk_function(
        description=(
            "Terminate the interaction with the open interpreter, resetting it to a fresh state. "
            "Whenever you finish a task with the open interpreter or no longer need it, you should promptly end the interaction with it."
        )
    )
    def end_open_interpreter(self):
        self.interpreter.reset()
        gptui_logger.info("Open interpreter reset.")
        return "Successfully terminated the interaction with the open interpreter."

"""



from typing import Any

from .driver_error import NoDriverError, NoDriverMethodError


class DriverInterface:

    def __init__(self, platform: str):
        self.platform = platform.lower()

    def __call__(self, *args, **kwargs) -> Any:
        method = getattr(self, self.platform, None)
        if method and callable(method):
            return method(*args, **kwargs)
        else:
            raise NoDriverError(self.platform)

    def termux(self):
        raise NoDriverMethodError(driver="termux", method=self.__class__.__name__)
    
    def linux(self):
        raise NoDriverMethodError(driver="linux", method=self.__class__.__name__)
    
    def macos(self):
        raise NoDriverMethodError(driver="macos", method=self.__class__.__name__)
    
    def windows(self):
        raise NoDriverMethodError(driver="windows", method=self.__class__.__name__)



from __future__ import annotations
import itertools
import logging
import os
import pyperclip
import subprocess
import threading
import time
from playsound import playsound
from threading import Thread
from typing import Callable, Self

from .driver_interface import DriverInterface
from ..models.utils.openai_api import openai_api_client


gptui_logger = logging.getLogger("gptui_logger")


class CopyCode(DriverInterface):
    
    def termux(self, content: str) -> None:
        subprocess.call(["termux-clipboard-set", content])

    def linux(self, content: str) -> None:
        pyperclip.copy(content)
    
    def macos(self, content: str) -> None:
        pyperclip.copy(content)


class TextToSpeak(DriverInterface):

    def __init__(self, dot_env_path: str, temp_dir: str, *args, **kwargs):
        self.openai_api_client = openai_api_client(dot_env_path=dot_env_path)
        if not os.path.exists(temp_dir):
            os.makedirs(temp_dir)
        self.temp_dir = temp_dir
        self._unique_id = itertools.count(1)
        self._audio_dict = {}
        self._now_audio_index = 1
        self._voice_service = None
        self._start_time = 0
        self._stop_time = 0
        self._play_audio_method: Callable = self._play_audio_termux
        self.cond = threading.Condition()
        super().__init__(*args, **kwargs)

    def _get_audio(self, text_content: str) -> str | None:
        send_time = time.time()
        now_id = next(self._unique_id)
        audio_file_name = f"speech_{now_id}.mp3"
        speech_file_path = os.path.join(self.temp_dir, audio_file_name)
        try:
            response = self.openai_api_client.audio.speech.create(
                model="tts-1",
                voice="nova",
                input=text_content,
            )
            response.stream_to_file(speech_file_path)
        except Exception as e:
            gptui_logger.error(e)
            with self.cond:
                self._now_audio_index += 1 # The current reading position scrolls forward by 1.
            return
        else:
            with self.cond:
                if self._stop_time > self._start_time: # Stop state
                    self._now_audio_index += 1 # The current reading position scrolls forward by 1.
                    return
                if send_time < self._stop_time:
                    # Eliminate messages sent before the stop but received after the start due to delay.
                    # That is, to exclude residual messages from before the start.
                    self._now_audio_index += 1 # The current reading position scrolls forward by 1.
                    return
                self._audio_dict[now_id] = speech_file_path
                if not self._voice_service:
                    voice_service = Thread(target=self._voice_speak)
                    voice_service.start()
                    self._voice_service = voice_service
                self.cond.notify_all()
        return speech_file_path
    
    def _voice_speak(self) -> None:
        while True:
            with self.cond:
                if self._now_audio_index in self._audio_dict:
                    audio_path = self._audio_dict.pop(self._now_audio_index)
                    self._now_audio_index += 1
                    speak = True
                else:
                    speak = False
                    status = self.cond.wait(timeout=5)
                    if not status:
                        self._voice_service = None
                        break
            if speak is True:
                self._play_audio_method(audio_path)
                os.remove(audio_path)
    
    def _play_audio_termux(self, audio_path: str) -> None:
        subprocess.run(
            ['termux-media-player', 'play', audio_path],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )
        while True:
            result = subprocess.run(['termux-media-player', 'info'], stdout=subprocess.PIPE, text=True)
            if "Playing" not in result.stdout:
                break
            time.sleep(0.1)

    def _play_audio_linux(self, audio_path: str) -> None:
        playsound(audio_path)

    def stop(self) -> None:
        self._stop_time = time.time()
        with self.cond:
            if self._voice_service is None:
                return
            else:
                for value in self._audio_dict.values():
                    self._now_audio_index += 1
                    os.remove(value)
                self._audio_dict = {}

    def termux(self, content: str) -> Self:
        self._start_time = time.time()
        self._play_audio_method = self._play_audio_termux
        Thread(target=self._get_audio, args=(content,)).start()
        return self

    def linux(self, content: str) -> Self:
        self._start_time = time.time()
        self._play_audio_method = self._play_audio_linux
        Thread(target=self._get_audio, args=(content,)).start()
        return self
    
    def macos(self, content: str) -> Self:
        return self.linux(content)


class VoiceRecordStart(DriverInterface):
    """Only supports recording in .wav format now."""

    def __init__(self, *args, **kwargs):
        self.record_flag = True
        super().__init__(*args, **kwargs)

    def termux(self, file_path: str, max_time: int) -> tuple[int, VoiceRecordStart]:
        if os.path.isfile(file_path):
            os.remove(file_path)

        _, file_extension = os.path.splitext(file_path)
        assert file_extension == ".wav", "Only supports recording in .wav format now."
        
        exit_code = subprocess.call(['termux-microphone-record', '-l', f'{max_time}', '-f', f'{file_path}', '-e', f'{file_extension[1:]}'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return exit_code, self

    def linux(self, file_path: str, max_time: int) -> tuple[int, VoiceRecordStart]:
        
        self.record_flag = True

        if os.path.isfile(file_path):
            os.remove(file_path)

        try:
            import pyaudio
            import wave
        except ImportError as e:
            gptui_logger.error(f"Import error: {e}")
            return 1, self

        CHUNK = 1024
        FORMAT = pyaudio.paInt16
        CHANNELS = 1
        RATE = 44100
        
        p = pyaudio.PyAudio()
        frames = []
        
        def record_audio():
            start_time = time.time()

            stream = p.open(format=FORMAT,
                            channels=CHANNELS,
                            rate=RATE,
                            input=True,
                            frames_per_buffer=CHUNK)

            while self.record_flag:
                elapsed_time = time.time() - start_time
                if elapsed_time > max_time:
                    break
                
                data = stream.read(CHUNK)
                frames.append(data)

            stream.stop_stream()
            stream.close()

            with wave.open(file_path, 'wb') as wf:
                wf.setnchannels(CHANNELS)
                wf.setsampwidth(p.get_sample_size(FORMAT))
                wf.setframerate(RATE)
                wf.writeframes(b''.join(frames))

            p.terminate()
        
        record_thread = threading.Thread(target=record_audio)
        record_thread.start()

        return 0, self

    def macos(self, file_path:str, max_time: int):
        return self.linux(file_path, max_time)


class VoiceRecordQuit(DriverInterface):

    def termux(self, _):
        exit_code = subprocess.call(['termux-microphone-record','-q'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return exit_code

    def linux(self, voice_record_start_handle: VoiceRecordStart):
        voice_record_start_handle.record_flag = False
        return 0
    
    def macos(self, voice_record_start_handle: VoiceRecordStart):
        voice_record_start_handle.record_flag = False
        return 0



import logging


class NullHandler(logging.Handler):
    def emit(self, record):
        pass


def get_null_logger(name=None):
    logger = logging.getLogger(name or __name__)
    logger.addHandler(NullHandler())
    logger.propagate = False
    return logger



from __future__ import annotations
import json
import logging
import sys
from abc import ABCMeta, abstractmethod
from importlib import import_module

import agere.commander as ac
import semantic_kernel as sk
from dotenv import dotenv_values
from semantic_kernel.connectors.ai.open_ai import OpenAITextCompletion
from semantic_kernel.orchestration.sk_context import SKContext
from semantic_kernel.orchestration.sk_function_base import SKFunctionBase

from .kernel_exceptions import (
    PluginInfoError,
    PluginsMatchError,
)
from .null_logger import get_null_logger


class KernelInterface(metaclass=ABCMeta):
    @property
    @abstractmethod
    def commander(self) -> ac.CommanderAsyncInterface:
        ...
    
    @commander.setter
    @abstractmethod
    def commander(self, commander: ac.CommanderAsyncInterface) -> None:
        ...
    
    @property
    @abstractmethod
    def sk_kernel(self) -> sk.Kernel:
        ...
    
    @sk_kernel.setter
    @abstractmethod
    def sk_kernel(self, sk_kernel: sk.Kernel) -> None:
        ...
    
    @abstractmethod
    def set_sk_kernel(self, sk_kernel: sk.Kernel, plugins_list: list[tuple]) -> None:
        ...
    
    @property
    @abstractmethod
    def functions_meta_in_kernel(self) -> dict:
        ...
    
    @property
    @abstractmethod
    def functions_link_in_kernel(self) -> dict:
        ...
    
    @property
    @abstractmethod
    def plugins_in_kernel(self) -> list[tuple]:
        ...
    
    @property
    @abstractmethod
    def llm_function_call(self) -> Kernel.LLM_Function_Call:
        ...

    @staticmethod
    @abstractmethod
    def make_basic_semantic_kernel(dot_env_config_path: str | None = None) -> sk.Kernel:
        ...

    @abstractmethod
    def make_commander(self, logger: logging.Logger | None = None) -> ac.CommanderAsyncInterface:
        ...

    @abstractmethod
    def add_plugins(self, plugins_list: list[tuple]) -> None:
        ...

    @abstractmethod
    def remove_plugins(self, plugins_list: list[tuple]) -> None:
        ...

    @abstractmethod
    def overwrite_plugins(self, plugins_list: list[tuple]) -> None:
        ...
    
    @abstractmethod
    def register_plugins(self, plugins_list: list[tuple]) -> None:
        ...

    @abstractmethod
    def context_render(self, args: dict, function: SKFunctionBase) -> SKContext:
        ...


class Kernel(KernelInterface):
    def __init__(self, dot_env_config_path: str | None = None, logger: logging.Logger | None = None):
        self.dot_env_config_path = dot_env_config_path
        self.logger = logger or get_null_logger()
        self._commander = self.make_commander(self.logger)
        self._sk_kernel = Kernel.make_basic_semantic_kernel(dot_env_config_path)
        self._llm_function_call = self.LLM_Function_Call(self)
        self._functions_meta_in_kernel = {}
        self._functions_link_in_kernel = {}
        self._plugins_in_kernel = []

    @property
    def commander(self) -> ac.CommanderAsyncInterface:
        return self._commander

    @commander.setter
    def commander(self, commander: ac.CommanderAsyncInterface) -> None:
        self._commander = commander

    @property
    def sk_kernel(self) -> sk.Kernel:
        return self._sk_kernel

    @sk_kernel.setter
    def sk_kernel(self, sk_kernel: sk.Kernel) -> None:
        try:
            self.set_sk_kernel(sk_kernel=sk_kernel, plugins_in_kernel=[])
        except PluginsMatchError:
            raise ValueError(
                "The kernel instance assigned to sk_kernel must be a kernel instance where the registered plugins are empty. "
                "If plugins are not empty, please use the set_sk_kernel method. "
                "It's recommended that all management of plugins be done through gk_kernel. "
            )

    def set_sk_kernel(self, sk_kernel: sk.Kernel, plugins_in_kernel: list[tuple] | None = None) -> None:
        plugins_in_kernel = plugins_in_kernel or []
        if self._dose_plugins_match_sk_kernel(sk_kernel=sk_kernel, plugins_list=plugins_in_kernel) is True:
            self._sk_kernel = sk_kernel
            self._plugins_in_kernel = plugins_in_kernel
        else:
            raise PluginsMatchError(sk_kernel, plugins_in_kernel)
    
    @property
    def llm_function_call(self) -> Kernel.LLM_Function_Call:
        return self._llm_function_call

    @property
    def functions_meta_in_kernel(self) -> dict:
        return self._functions_meta_in_kernel

    @property
    def functions_link_in_kernel(self) -> dict:
        return self._functions_link_in_kernel

    @property
    def plugins_in_kernel(self) -> list[tuple]:
        if self._dose_plugins_match_sk_kernel(sk_kernel=self.sk_kernel, plugins_list=self._plugins_in_kernel) is True:
            return self._plugins_in_kernel
        else:
            raise PluginsMatchError(self.sk_kernel, self._plugins_in_kernel)

    @staticmethod
    def make_basic_semantic_kernel(dot_env_config_path: str | None = None) -> sk.Kernel:
        dot_env_config_path = dot_env_config_path or '.env'
        dot_env_config = dotenv_values(dot_env_config_path)
        openai_key = dot_env_config.get("OPENAI_API_KEY", None)
        org_id = dot_env_config.get("OPENAI_ORG_ID", None)
        assert openai_key, f"OpenAI API key not found in {dot_env_config_path} file"
        kernel = sk.Kernel()
        kernel.add_text_completion_service(               # We are adding a text service
            "OpenAI_davinci",                         # The alias we can use in prompt templates' config.json
            OpenAITextCompletion(
                "text-davinci-003",                   # OpenAI Model Name
                openai_key,
                org_id,
            )
        )
        kernel.set_default_text_completion_service("OpenAI_davinci")
        return kernel

    def make_commander(self, logger: logging.Logger | None = None) -> ac.CommanderAsyncInterface:
        return ac.CommanderAsync(logger or get_null_logger())

    def add_plugins(self, plugins_list: list[tuple]) -> None:
        self.register_plugins(plugins_list)
    
    def remove_plugins(self, plugins_list: list[tuple]) -> None:
        new_plugins_list = [plugin for plugin in self.plugins_in_kernel if plugin not in plugins_list]
        self.set_sk_kernel(self.make_basic_semantic_kernel(self.dot_env_config_path))
        self.register_plugins(new_plugins_list)

    def overwrite_plugins(self, plugins_list: list[tuple]) -> None:
        self.set_sk_kernel(self.make_basic_semantic_kernel(self.dot_env_config_path))
        self.register_plugins(plugins_list)

    def register_plugins(
        self,
        plugins_list: list[tuple],
    ) -> None:
        """
        args:
            plugins_list: plugins to include
            [
                (plugin_path, plugin_name) for semantic plugin,
                (plugin_path, plugin_file, plugin_name) for natice plugin,
                    - it can also accept the initialization parameters of the class in the form of a tuple and a dictionary as the fourth and fifth optional parameters.
            ]
        """
        new_plugins_list = [plugin for plugin in plugins_list if plugin not in self.plugins_in_kernel]
        _, functions_meta_in_kernel, functions_link_in_kernel = register_plugins_to_sk_kernel(
            plugins_list=new_plugins_list,
            sk_kernel=self.sk_kernel,
        )

        # Refresh plugins_in_kernel, functions_link_in_kernel and functions_meta_in_kernel
        self._plugins_in_kernel.extend(new_plugins_list)
        self._functions_meta_in_kernel = functions_meta_in_kernel
        self._functions_link_in_kernel = functions_link_in_kernel

    def context_render(self, args: dict, function: SKFunctionBase) -> SKContext:
        context = self.sk_kernel.create_new_context()
        params = function.describe().parameters
        for param in params:
            name = param.name
            default_value = param.default_value
            context[name] = default_value
        for arg, value in args.items():
            context[arg] = value
        return context

    def _dose_plugins_match_sk_kernel(self, sk_kernel: sk.Kernel, plugins_list: list[tuple]) -> bool:
        """
        Check if the plugins in plugins_list is the same as the plugins in the kernel.
        """
        # Get all plugins' names in self.sk_kernel
        native_functions = sk_kernel.skills.get_functions_view().native_functions
        semantic_functions = sk_kernel.skills.get_functions_view().semantic_functions
        native_functions.update(semantic_functions)
        all_functions = native_functions
        skill_names = set(all_functions.keys())
        # Get all plugins' names in plugins_list
        plugin_names = set()
        for plugin in plugins_list:
            if len(plugin) == 2:
                plugin_names.add(plugin[1])
            elif len(plugin) >= 3:
                plugin_names.add(plugin[2])
        if plugin_names == skill_names:
            return True
        else:
            return False

    class LLM_Function_Call:
        def __init__(self, kenel: Kernel):
            self.kenel = kenel
        
        def sk_plugins_to_openai_functions(self, 
            plugins_list: list = [], 
            functions: list = [], 
            to_user: bool = True,
            auto_remove_input: bool = True
        ) -> tuple[list[dict], dict]:
            """Retrieve the meta and links of the specified function registered in the kernel for use in LLM function calls.
            If both plugins_list and functions are not specified, all functions registered in the kenel will be returned.
            
            Args:
                plugins_list: Add specified plugins
                functions: Add specified functions
                to_user: Add to user ability
                auto_remove_input: Automatically remove the 'input' arg from the args list, which is automatically added by 'sk_kernel'.
                    When the arg name has been specified, this 'input' arg is actually redundant and useless.
                    The conditions for automatically removing 'input' arg is:
                        There are more than two args in the args list.
                        The description of 'input' is empty.
            
            Return:
                (
                    available_functions_meta: list,
                    available_functions_link: dict,
                )
            """
            functions_meta = []
            all_functions_meta_in_kernel = self.kenel.functions_meta_in_kernel
            if plugins_list or functions:
                for function in all_functions_meta_in_kernel.values():
                    if isinstance(function, dict):
                        # function is not organized by plugins
                        if (function["parent"] in plugins_list) or (function["name"] in functions):
                            functions_meta.append(function)
                    elif isinstance(function, list):
                        # function is organized by plugins
                        function_list = function
                        for one_function_dict in function_list:
                            function = list(one_function_dict.values())[0]
                            if (function["parent"] in plugins_list) or (function["name"] in functions):
                                functions_meta.append(function)
            else:
                for function in all_functions_meta_in_kernel.values():
                    if isinstance(function, dict):
                        # function is not organized by plugins
                        functions_meta.append(function)
                    elif isinstance(function, list):
                        # function is organized by plugins
                        function_list = function
                        for one_function_dict in function_list:
                            function = list(one_function_dict.values())[0]
                            functions_meta.append(function)
            
            available_functions_meta = []
            available_functions_link = {}
            for function in functions_meta:
                available_functions_meta.append(
                    {
                        "type": "function",
                        "function": {
                            "name": function["name"],
                            "description": function["description"],
                            "parameters": {
                                "type": "object",
                                "properties": (parameters := {}),
                                "required": (required := []),
                            },
                        },
                    }
                )
                if to_user is True:
                    parameters["to_user"] = {
                        "type": "string",
                        "description": "The content paraphrased for the user. The content of this parameter can tell the user what you are about to do, or it can be an explanation of the behavior of the function calling. For example, 'I'm going to search the internet, please wait a moment.'",
                        }
                params_num = len(function["args"])
                for param_name, param in function["args"].items():
                    if auto_remove_input is True and params_num >= 2:
                        if param_name == "input" and not param["description"]:
                            continue
                    parameters[param_name] = {
                        "type": "string",
                        "description": param["description"]
                    }
                    if param["default_value"] is None:
                        required.append(param_name)

                available_functions_link[function["name"]] = self.kenel.functions_link_in_kernel[function["name"]]
                
            return available_functions_meta, available_functions_link


class PluginMeta:
    def __init__(self, plugin_info: tuple):
        self.plugin_info = plugin_info

    def is_semantic(self):
        if len(self.plugin_info) == 2:
            return True
        else:
            return False

    def is_native(self):
        if len(self.plugin_info) >= 3:
            return True
        else:
            return False

    @property
    def name(self) -> str:
        return self.plugin_info[1] if len(self.plugin_info) == 2 else self.plugin_info[2]
    
    @property
    def functions_list(self) -> list:
        out = []
        meta = self.functions_meta
        for _, value in meta.items():
            out.append(value["name"])
        return out

    @property
    def functions_meta(self) -> dict:
        meta = get_plugins_functions_info(plugins_list=[self.plugin_info])
        return meta


def register_plugins_to_sk_kernel(
    sk_kernel: sk.Kernel,
    plugins_list: list[tuple],
) -> tuple:
    """
    args:
        sk_kernel: register to this kernel
        plugins_list: plugins to include
            [
                (plugin_path, plugin_name) for semantic plugin,
                (plugin_path, plugin_file, plugin_name) for natice plugin:
                    - it can also accept the initialization parameters of the class in the form of a tuple and a dictionary as the fourth and fifth optional parameters.
            ]
    retrun:
        (
            sk_kernel: semantic_kernel.Kernel
            functions_meta_in_kernel: dict
            functions_link_in_kernel: dict
        )
    """

    with TemporarySysPath():
        # register plugin
        for plugin_info in plugins_list:
            if len(plugin_info) == 2:
                sk_kernel.import_semantic_skill_from_directory(plugin_info[0], plugin_info[1])
            elif len(plugin_info) >= 3:
                if plugin_info[0] not in sys.path:
                    sys.path.append(plugin_info[0])
                plugin_file = import_module(plugin_info[1])
                plugin = getattr(plugin_file, plugin_info[2])
                args = ()
                kwargs = {}
                if len(plugin_info) >= 4:
                    params = plugin_info[3:5]
                    for param in params:
                        if isinstance(param, tuple):
                            args = param
                        elif isinstance(param, dict):
                            kwargs = param
                        else:
                            raise PluginInfoError(plugin_info)
                sk_kernel.import_skill(plugin(*args, **kwargs), plugin_info[2])
            else:
                raise PluginInfoError(plugin_info)
    
    functions_meta_in_sk_kernel = get_functions_meta_in_sk_kernel(sk_kernel=sk_kernel)
    functions_link_in_sk_kernel = get_functions_link_in_sk_kernel(sk_kernel=sk_kernel)
    
    return sk_kernel, functions_meta_in_sk_kernel, functions_link_in_sk_kernel

def get_plugins_functions_info(
    plugins_list: list[tuple],
    functions: list | None = None,
    organize_by_plugins: bool = False,
) -> dict:
    """
    Retrieve functions infomation from plugins, it does not register these plugins into the existing kernel, but creates a new temporary kernel to handle this information.
    args:
        plugins_list {list[tuple]}: plugins to include
            [
                (plugin_path, plugin_name) for semantic plugin,
                (plugin_path, plugin_file, plugin_name) for natice plugin,
                    - it can also accept the initialization parameters of the class in the form of a tuple and a dictionary as the fourth and fifth optional parameters.
            ]
        functions {list | None}: function filter, only function in funtions is picked if it is specified
        organize_by_plugins {bool}: functions meta are organized by plugins
    retrun:
        functions_meta: dict
    """
    sk_kernel = sk.Kernel()
    result_tuple = register_plugins_to_sk_kernel(plugins_list=plugins_list, sk_kernel=sk_kernel)
    if functions is None and organize_by_plugins is False:
        return result_tuple[1]
    else:
        functions_meta = get_functions_meta_in_sk_kernel(sk_kernel=result_tuple[0], functions=functions, organize_by_plugins=organize_by_plugins)
        return functions_meta

def get_functions_meta_in_sk_kernel(
    sk_kernel: sk.Kernel,
    functions: list | None = None,
    organize_by_plugins: bool = False,
) -> dict:
    """
    args:
        sk_kernel: get available functions in this kernel
        functions: function filter, only function in funtions is picked if it is specified
        organize_by_plugins: functions are organized by plugins
    return:
        organize_by_plugins = False:
            {
                plugin_name.function_name: {    # plugin_name in key is necessary, because it's possible to have a same name function in different plugins
                    "name": function_name,
                    "description": function_description
                    "args": {
                        arg_name: {
                            "name": param_name
                            "description": param_description
                            "default_value": param_default_value ---- None for no default_value, means it is a necessary parameter. if default_value is None, it should be a str "None".
                        }
                    "parent": plugin
                    }
                }
            }
        organize_by_plugins = True:
            {
                plugin_name: {
                    [
                        {
                            plugin_name.function_name: {
                                "name": function_name,
                                "description": function_description
                                "args": {
                                    arg_name: {
                                        "name": param_name
                                        "description": param_description
                                        "default_value": param_default_value ---- None for no default_value, means it is a necessary parameter. if default_value is None, it should be a str "None".
                                    }
                                "parent": plugin
                                }
                            }
                        },
                    ]
                }
            }
    """

    kernel = sk_kernel

    # Get a dictionary of skill names to all native and semantic functions
    native_functions = kernel.skills.get_functions_view().native_functions
    semantic_functions = kernel.skills.get_functions_view().semantic_functions
    native_functions.update(semantic_functions)
    all_functions = native_functions

    # Create a mapping between all function names and their descriptions
    # and also a mapping between function names and their parameters
    skill_names = list(all_functions.keys())
    all_functions_descriptions_dict = {}
    all_functions_params_dict = {}
    all_functions_skill_dict = {}
    all_functions_name_dict = {}
    plugin_dict = {}

    for skill_name in skill_names:
        for func in all_functions[skill_name]:
            if functions is not None:
                if func.name not in functions:
                    continue
            key = skill_name + "." + func.name
            all_functions_descriptions_dict[key] = func.description
            all_functions_params_dict[key] = func.parameters
            all_functions_skill_dict[key] = skill_name
            all_functions_name_dict[key] = func.name

    # Create the [AVAILABLE FUNCTIONS] section of the prompt
    available_functions = {}
    for name in list(all_functions_descriptions_dict.keys()):
        available_functions[name] = {
            "name":all_functions_name_dict[name],
            "description":all_functions_descriptions_dict[name],
            "args":{},
            "parent":all_functions_skill_dict[name]
        }

        # Add the parameters for each function
        parameters = all_functions_params_dict[name]
        for param in parameters:
            if not param.description:
                param_description = ""
            else:
                param_description = param.description
            if not param.default_value:
                param_default_value = None
            else:
                param_default_value = param.default_value

            available_functions[name]["args"][param.name] = {}
            available_functions[name]["args"][param.name]["name"] = param.name
            available_functions[name]["args"][param.name]["description"] = param_description
            available_functions[name]["args"][param.name]["default_value"] = param_default_value
    
        if organize_by_plugins is True:
            skill_name = all_functions_skill_dict[name]
            if skill_name not in plugin_dict:
                plugin_dict[skill_name] = []
            plugin_dict[skill_name].append({name:available_functions[name]})
            
    if organize_by_plugins is True:
        return plugin_dict
    else:
        return available_functions

def get_functions_link_in_sk_kernel(sk_kernel: sk.Kernel, functions: list | None = None) -> dict:
    """
    args:
        sk_kernel: get available functions' links in this kernel
        functions: function filter, only function in funtions is picked if it is specified
    return:
        {
            function_name: function,
        }
    """

    # Get a dictionary of skill names to all native and semantic functions
    native_functions = sk_kernel.skills.get_functions_view().native_functions
    semantic_functions = sk_kernel.skills.get_functions_view().semantic_functions
    native_functions.update(semantic_functions)
    all_functions = native_functions

    # Create a mapping between all function names and functions
    skill_names = list(all_functions.keys())
    functions_link = {}

    for skill_name in skill_names:
        for func in all_functions[skill_name]:
            if functions is not None:
                if func.name not in functions:
                    continue
            key = func.name
            function_obj = sk_kernel.skills.get_function(skill_name, func.name)
            functions_link[key] = function_obj
    return functions_link

def to_json_string(input, ensure_ascii: bool = False, sort_keys: bool = True, indent: int = 4, separators = (',',':'), **kwargs):
    if not isinstance(input, str):
        input = json.dumps(input, ensure_ascii=ensure_ascii, sort_keys=sort_keys, indent=indent, separators=separators, **kwargs)
    return input
    
def pretty_dict(dict_object: dict|None) -> str:
    if dict_object is None:
        return ""
    return json.dumps(dict_object, ensure_ascii = False, sort_keys = True, indent = 4, separators = (',',':'))


class TemporarySysPath:
    def __enter__(self):
        self.original_sys_path = sys.path[:]
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.path = self.original_sys_path



import logging

from ..models.signals import response_to_user_message_sentence_stream_signal


gptui_logger = logging.getLogger("gptui_logger")


class VoiceService:
    def __init__(self, myapp, switch: bool = True):
        self.myapp = myapp
        self.voice_service = None
        if switch is True:
            self.connect()

    async def accept_voice_message(self, sender, **kwargs):
        voice_message = kwargs["message"]
        message_content = voice_message["content"]
        flag = voice_message["flag"]
        if flag == "content":
            self.voice_service = self.myapp.drivers.tts(message_content)

    def connect(self):
        response_to_user_message_sentence_stream_signal.connect(self.accept_voice_message)

    def disconnect(self):
        response_to_user_message_sentence_stream_signal.disconnect(self.accept_voice_message)

    def cancel_speak(self) -> None:
        self.disconnect()
        if self.voice_service is None:
            return
        self.voice_service.stop()



import logging

from .my_text import MyText as Text
from ..views.theme import theme_color as tc

gptui_logger = logging.getLogger("gptui_logger")


def file_icon(
    file_label: str,
    file_type: str,
    file_description: str,
    icon_color: str | None = None,
    description_color: str | None = None,
) -> Text:
    icon_color = icon_color or tc("yellow") or "yellow"
    description_color = description_color or tc("white") or "white"

    display = Text('', icon_color)
    if file_type == ".txt":
        display += Text('\u2595'+'\u2056\u0305'+'\u2056\u0305'+'\u2056\u0305'+'\u2572'+' \n')
        display += Text('\u2595'+f'{file_label[:3]}'+'\u2595'+' \n'+'\u2595')
        display += Text('txt\u2595', 'underline')
    elif file_type == ".md":
        display += Text('\u2595'+' \u0305'+'\ueb1d\u0305'+' \u0305'+'\u2572'+' \n')
        display += Text('\u2595'+f'{file_label[:3]}'+'\u2595'+' \n'+'\u2595')
        display += Text('.md\u2595', 'underline')
    elif file_type == ".bin":
        display += Text('\u2595'+'l\u0305'+'l\u0305'+'l\u0305'+'l\u0305'+'l\u0305'+'l\u0305'+'l\u0305'+'\u2572'+' \n')
        display += Text('\u2595'+f'{file_label[:3]}'+'\u2595'+' \n'+'\u2595')
        display += Text('bin\u2595', 'underline')
    elif file_type == ".json":
        display += Text('\u2595'+' \u0305'+'{\u0305'+' \u0305'+'}\u0305'+'\u2572'+' \n')
        display += Text('\u2595'+f'{file_label[:3]}'+'\u2595'+' \n'+'\u2595')
        display += Text('jsn\u2595', 'underline')
    elif file_type == ".py":
        display += Text('\u2595'+' \u0305'+'\ue606\u0305'+' \u0305'+'\u2572'+' \n')
        display += Text('\u2595'+f'{file_label[:3]}'+'\u2595'+' \n'+'\u2595')
        display += Text('.\uf820 \u2595', 'underline')
    elif file_type == ".sh":
        display += Text('\u2595'+'<\u0305'+'\u29f8\u0305'+'>\u0305'+'\u2572'+' \n')
        display += Text('\u2595'+f'{file_label[:3]}'+'\u2595'+' \n'+'\u2595')
        display += Text('.sh\u2595', 'underline')
    else:
        file_type += '   '
        display += Text('\u2595'+'\u203e'+'\u203e'+'\u203e'+'\u29f9'+' \n')
        display += Text('\u2595'+f'{file_label[:3]}'+'\u2595'+' \n'+'\u2595')
        display += Text(f'{file_type[:3]}\u2595', 'underline')
    if len(file_description) > 12:
        description_line0 = file_description[:6] + '\n'
        description_line1 = '\u2026' + file_description[-5:] + '\n'
    else:
        file_description = file_description.ljust(12)
        description_line0 = file_description[:6] + '\n'
        description_line1 = file_description[6:] + '\n'
    description = Text(' \n' + description_line0 + description_line1, f'{description_color}')
    out_display = display + description
    return out_display



import logging

from ..models.signals import common_message_signal


gptui_logger = logging.getLogger("gptui_logger")


class GroupTalkControl:
    def __init__(self, app):
        self.app = app
        common_message_signal.connect(self.group_talk_user_message_send)

    def group_talk_user_message_send(self, sender, **kwargs):
        message = kwargs["message"]
        if message["flag"] == "group_talk_user_message_send":
            messages = message["content"]
            if isinstance(messages, list):
                for one_message in messages:
                    self.app.context_piece_to_chat_window(one_message, change_line=True, decorator_switch=True)
            else:
                self.app.context_piece_to_chat_window(messages, change_line=True, decorator_switch=True)



import logging
import json
import threading
import time

from agere.commander import Callback
from semantic_kernel.orchestration.sk_context import SKContext
from semantic_kernel.skill_definition import sk_function, sk_function_context_parameter

from gptui.gptui_kernel.manager import ManagerInterface, auto_init_params
from gptui.models.blinker_wrapper import async_wrapper_with_loop, sync_wrapper
from gptui.models.openai_chat_inner_service import chat_service_for_inner
from gptui.models.openai_error import OpenaiErrorHandler
from gptui.models.signals import response_auxiliary_message_signal, notification_signal
from gptui.models.utils.openai_api import openai_api
from gptui.utils.my_text import MyText as Text

gptui_logger = logging.getLogger("gptui_logger")


class SnoozeReminder:
    def __init__(self, manager: ManagerInterface):
        self.manager = manager
        self.openai_api = openai_api(manager.dot_env_config_path)
    
    @auto_init_params("0")
    @classmethod
    def get_init_params(cls, manager) -> tuple:
        return (manager,)

    @sk_function(
        description="Set a reminder. When the time comes, you will be notified of the content you set.",
        name="snooze_reminder",
    )
    @sk_function_context_parameter(
        name="delay",
        description="Set the delay for the reminder, in seconds.",
    )
    @sk_function_context_parameter(
        name="reminder_content",
        description="Content to be reminded."
    )
    @sk_function_context_parameter(
        name="openai_context",
        description=(
            "The dictionary string version of the OpenaiContext instance. "
            "This is a special parameter that typically doesn't require manual intervention, as it is usually automatically managed."
            "Unless there's a clear intention, please keep its default value."
        ),
        default_value="AUTO"
    )
    def snooze_reminder(self, context: SKContext) -> str:
        delay = context["delay"]
        
        try:
            delay = int(delay)
        except ValueError:
            return "The parameter 'delay' cannot be parsed into integer seconds."

        content = context["reminder_content"]
        openai_context_dict = json.loads(str(context["openai_context"]))
        conversation_id = openai_context_dict["id"]
        
        reminder = threading.Timer(delay, self.reminder_after, args=(content, conversation_id, delay))
        reminder.start()
        return "The reminder has been set."

    def reminder_after(self, content: str, conversation_id: int | str, delay: int) -> None:
        content = f"You(!not the user!) set a reminder {str(delay)} seconds ago, and the {str(delay)} seconds have elapsed. The time is up.\n==========REMINDER CONTENT BEGIN==========\n" + content
        content += "\n==========REMINDER CONTENT END==========\nThis message is a reminder for you, not for the user. You should handle this message appropriately based on the conversation history."
        openai_chat_manager = self.manager.client.openai
        conversation = openai_chat_manager.conversation_dict.get(conversation_id)
        if conversation is None:
            self.manager.client.query_one("#status_region").update(Text("You have a reminder whose time has come, but it seems the conversation has been closed.", "yellow"))
            gptui_logger.info("You have a reminder whose time has come, but it seems the conversation has been closed.")
        openai_context = conversation["openai_context"]
        self.manager.client.query_one("#chat_tabs").active = "lqt" + str(conversation_id)
        time.sleep(1)
        functions = self.manager.available_functions_meta
        messages_list=[
            {
                "role": "function",
                "name": "snooze_reminder",
                "content": content,
            }
        ]
        response_auxiliary_message_signal.send(
            self,
            message={
                "content":{
                        "role": "function",
                        "name": "snooze_reminder",
                        "content": content,
                },
                "flag": "function_call",
            }
        )

        try:
            response = chat_service_for_inner(
                messages_list=messages_list,
                context=openai_context,
                openai_api=self.openai_api,
                functions=functions,
                function_call="auto",
            )
        except Exception as e:
            OpenaiErrorHandler().openai_error_handle(error=e, context=openai_context)
        
        openai_chat_manager.openai_chat.chat_messages_extend(messages_list=messages_list, context=openai_context)
        
        ResponseJob = self.manager.get_job("ResponseJob")
        callback = Callback(
            at_job_start=[
                {
                    "function": notification_signal.send,
                    "params": {
                        "args": (self,),
                        "kwargs": {
                            "_async_wrapper": async_wrapper_with_loop,
                            "message":{
                                "content":{
                                    "content":{"status":True},
                                    "description":"Commander status changed",
                                },
                                "flag":"info",
                            },
                        },
                    },
                },
            ],
            at_commander_end=[
                {
                    "function": notification_signal.send_async,
                    "params": {
                        "args": (self,),
                        "kwargs": {
                            "_sync_wrapper": sync_wrapper,
                            "message":{
                                "content":{
                                    "content":{"status":False},
                                    "description":"Commander status changed",
                                },
                                "flag":"info",
                            },
                        },
                    },
                },
            ],
        )
        job = ResponseJob(manager=self.manager, response=response, context=openai_context, callback=callback)
        self.manager.gk_kernel.commander.async_commander_run(job)



from __future__ import annotations
import inspect
import math
import logging
import time
import weakref
from abc import ABCMeta, abstractmethod
from typing import Literal, Iterable
from threading import Thread

from textual.message import Message

from ..utils.my_text import MyText as Text


gptui_logger = logging.getLogger("gptui_logger")


class Animation(metaclass=ABCMeta):
    animation_name: str
    @abstractmethod
    def run(self, animation_instance: AnimationManager.AnimationThread, message: AnimationRequest) -> None:
        ...


class AnimationRequest(Message):
    def __init__(
        self,
        ani_id,
        action: Literal["start", "end"],
        displayer: str = "default",
        priority: int = 2,
        ani_speed: float = 0.1,
        ani_type: str = "default",
        keep_time: float = 0,
        ani_end_display: str | Text = "recovery",
        others = None,
    ) -> None:
        """
        ani_end_display can be string or Text instance;
        if ani_end_display equals "default", it will display default content setted by AnimationManager
        if ani_end_display equals "recovery", it will display the erased content when animation started
        if others, it will be displayed directly.
        others for extended functions
        """
        self.ani_id = ani_id
        self.action = action
        self.displayer = displayer
        self.priority = priority
        self.ani_speed = ani_speed
        self.ani_type = ani_type
        self.keep_time = keep_time
        self.ani_end_display = ani_end_display
        self.others = others
        super().__init__()


def find_animations():
    current_module = inspect.getmodule(inspect.currentframe())
    animations = {}
    for _, obj in inspect.getmembers(current_module, predicate=inspect.isclass):
        if issubclass(obj, Animation) and not inspect.isabstract(obj):
            animations[obj.animation_name] = obj
    return animations


class AnimationManager:
    def __init__(self, displayer: dict, ani_end_display: str | Text = '', ani_links: dict | None = None):
        self.ani_id_dict = weakref.WeakValueDictionary()
        self.priority_dict = {}
        self.ani_links = ani_links or find_animations()
        self.displayer = displayer
        for key, _ in displayer.items():
            self.priority_dict[key] = weakref.WeakKeyDictionary()
        self.ani_end_display = ani_end_display

    def animation_start(self, message: AnimationRequest):
        animation_instance = self.AnimationThread(animation_manager=self, message=message)
        self.ani_id_dict[message.ani_id] = animation_instance
        animation_instance.start()

    def animation_end(self, ani_id) -> None:
        ani = self.ani_id_dict.get(ani_id)
        if ani:
            ani.end()

    def manage(self, message: AnimationRequest):
        if message.action == "start":
            self.animation_start(message)
        elif message.action == "end":
            self.animation_end(ani_id=message.ani_id)
    

    class AnimationThread(Thread):
        def __init__(self, animation_manager: AnimationManager, message: AnimationRequest) -> None:
            super().__init__()
            self.animation_status = True
            self.animation_manager = animation_manager
            self.message = message
            animation_manager.priority_dict[message.displayer][self] = message.priority

        def run(self) -> None:
            animation = self.animation_manager.ani_links.get(self.message.ani_type)
            if animation is None:
                raise ValueError(f"AnimationManager received a request for an animation type with no corresponding match. ani_type: {self.message.ani_type}")
            assert isinstance(animation, type(Animation))
            animation().run(animation_instance=self, message=self.message)

        def end(self) -> None:
            self.animation_status = False


def play_animation(
    animation_instance: AnimationManager.AnimationThread,
    message: AnimationRequest,
    keep_time,
    frames: Iterable,
    ) -> None:
    time_start = time.time()
    animation_manager = animation_instance.animation_manager
    display_object = animation_manager.displayer[message.displayer]
    if message.ani_end_display == "recovery":
        try:
            end_display = display_object.renderable
        except:
            end_display = animation_manager.ani_end_display
    elif message.ani_end_display == "default":
        end_display = animation_manager.ani_end_display
    else:
        end_display = message.ani_end_display
    while True:
        for frame in frames:
            if not animation_instance.animation_status or time.time() - time_start >= keep_time:
                display_object.update(end_display)
                return
            priority_list = animation_manager.priority_dict[message.displayer].values()
            if any(priority < message.priority for priority in priority_list):
                continue
            display_object.update(frame)
            time.sleep(message.ani_speed)


class DefaultAnimation(Animation):
    
    animation_name = "default"
    
    def run(
        self,
        animation_instance: AnimationManager.AnimationThread,
        message: AnimationRequest,
    ) -> None:
        frames_list = ["-- -- -- -- -- -- -- -- -- -- ", " -- -- -- -- -- -- -- -- -- --", "- -- -- -- -- -- -- -- -- -- -"]
        keep_time = message.keep_time or math.inf
        play_animation(
            animation_instance=animation_instance,
            message=message,
            keep_time=keep_time,
            frames=frames_list,
        )


class DefaultAnimation2(Animation):
    
    animation_name = "default_2"
    
    def run(
        self,
        animation_instance: AnimationManager.AnimationThread,
        message: AnimationRequest,
    ) -> None:
        frames_list = [
            "// // // // // // // // // // ",
            "// // // // // // // // // // ",
            " // // // // // // // // // //",
            " // // // // // // // // // //",
            "/ // // // // // // // // // /",
            "/ // // // // // // // // // /",
        ]
        keep_time = message.keep_time or math.inf
        play_animation(
            animation_instance=animation_instance,
            message=message,
            keep_time=keep_time,
            frames=frames_list,
        )


class DefaultAnimation3(Animation):
    
    animation_name = "default_3"
    
    def run(
        self,
        animation_instance: AnimationManager.AnimationThread,
        message: AnimationRequest,
    ) -> None:
        frames_list = [
            "|  |  |  |  |  |  |  |  |  |",
            " / |  |  |  |  |  |  |  |  |",
            "  __/ |  |  |  |  |  |  |  |",
            "  __ __/ |  |  |  |  |  |  |",
            "  __ __ __/ |  |  |  |  |  |",
            "  __ __ __ __/ |  |  |  |  |",
            "  __ __ __ __ __/ |  |  |  |",
            "  __ __ __ __ __ __/ |  |  |",
            "  __ __ __ __ __ __ __/ |  |",
            "  __ __ __ __ __ __ __ __/ |",
            "  __ __ __ __ __ __ __ __ __",
        ]
        keep_time = message.keep_time or math.inf
        play_animation(
            animation_instance=animation_instance,
            message=message,
            keep_time=keep_time,
            frames=frames_list,
        )


class StaticDisplayAnimation(Animation):
    
    animation_name = "static"
    
    def run(
        self,
        animation_instance: AnimationManager.AnimationThread,
        message: AnimationRequest,
    ) -> None:
        static_content = message.others
        keep_time = message.keep_time or 0
        frames_list = [static_content]
        play_animation(
            animation_instance=animation_instance,
            message=message,
            keep_time=keep_time,
            frames=frames_list,
        )


class SettingMemoryAnimation(Animation):
    
    animation_name = "setting_memory"
    
    def run(
        self,
        animation_instance: AnimationManager.AnimationThread,
        message: AnimationRequest,
    ) -> None:
        frames_list = [
                Text("Setting memory... -", "yellow"),
                Text("Setting memory... \\", "yellow"),
                Text("Setting memory... |", "yellow"),
                Text("Setting memory... /", "yellow"),
                ]
        play_animation(
            animation_instance=animation_instance,
            message=message,
            keep_time=math.inf,
            frames=frames_list,
        )



from __future__ import annotations
import logging
from abc import ABCMeta, abstractmethod
from typing import Literal

from agere.commander import Callback
from openai.types.chat import ChatCompletionMessageParam

from .blinker_wrapper import async_wrapper_without_loop, async_wrapper_with_loop, sync_wrapper
from .context import OpenaiContext
from .jobs import ResponseJob, GroupTalkManager
from .openai_error import OpenaiErrorHandler
from .openai_tokens_truncate import trim_excess_tokens
from .signals import notification_signal, chat_context_extend_for_sending_signal
from .utils.openai_api import openai_api_client
from .utils.tokens_num import tokens_num_for_functions_call
from ..gptui_kernel.manager import ManagerInterface


gptui_logger = logging.getLogger("gptui_logger")


class OpenaiChatInterface(metaclass=ABCMeta):
    @abstractmethod
    def chat_message_append(self, context: OpenaiContext, message: dict | list[dict]) -> None:
        ...
    
    @abstractmethod
    def chat_messages_extend(self, context: OpenaiContext, messages_list: list[dict]) -> None:
        ...
    
    @abstractmethod
    def chat_message_pop(self, context: OpenaiContext, pop_index: int = -1) -> dict:
        ...
    
    @abstractmethod
    def chat(self, context: OpenaiContext, message: dict | list[dict]) -> None:
        ...
    
    @abstractmethod
    def chat_stream(self, context: OpenaiContext, message: dict | list[dict]) -> None:
        ...


class OpenaiChat(OpenaiChatInterface):
    def __init__(self, manager: ManagerInterface):
        self.manager = manager
        self.openai_api_client = openai_api_client(manager.dot_env_config_path)

    def chat_message_append(self, context: OpenaiContext, message: dict | list[dict], tokens_num_update: bool = True) -> None:
        "Append chat message to the end of the chat_context of the context"
        if isinstance(message, dict):
            context.chat_context_append(message=message, tokens_num_update=tokens_num_update)
        else:
            for one_message in message:
                context.chat_context_append(message=one_message, tokens_num_update=tokens_num_update)

    def chat_messages_extend(self, context: OpenaiContext, messages_list: list[dict], tokens_num_update: bool = True) -> None:
        "Append a list of chat message to the end of the chat_context of the context"
        for message in messages_list:
            context.chat_context_append(message=message, tokens_num_update=tokens_num_update)
    
    def chat_message_pop(self, context: OpenaiContext, pop_index: int = -1) -> ChatCompletionMessageParam:
        "Pop a message from context"
        if context.chat_context is None:
            raise ValueError(f"Field 'chat_context' in {context.chat_context} has not been set.")
        return context.chat_context.pop(pop_index)

    def chat(self, context: OpenaiContext, message: dict | list[dict]) -> None:
        "Chat with openai"
        if isinstance(message, dict):
            messages_list = [message]
        else:
            messages_list = message

        if context.chat_context_saver_for_sending == "outer":
            chat_context_extend_for_sending_signal.send(
                self,
                _async_wrapper=async_wrapper_without_loop,
                message={
                    "content":{
                        "messages":messages_list,
                        "context":context,
                    },
                    "flag":"",
                }
            )
        else:
            self.chat_message_append(context=context, message=message)

        notification_signal.send(
            self,
            _async_wrapper=async_wrapper_without_loop,
            message={
                "content":{
                    "content":{"context":context},
                    "description":"Starting to send the original chat message from the user.",
                },
                "flag":"info",
            }
        )
        try:
            tools_para = {}
            if available_functions := self.manager.available_functions_meta:
                tools_para = {"tools": available_functions, "tool_choice": "auto"}
                response_mode = "function_call"
            else:
                response_mode = "no_function_call"
            
            offset_tokens_num = -tokens_num_for_functions_call(tools_para["tools"], model=context.parameters["model"])
            trim_messages = trim_excess_tokens(context, offset=offset_tokens_num)

            # Delete the tool reply messages at the beginning of the information list.
            # This is because if the information starts with a function reply message,
            # it indicates that the function call information has already been truncated.
            # The OpenAI API requires that function reply messages must be responses to function calls.
            # Therefore, if the function reply messages are not removed, it will result in an OpenAI API error.
            while trim_messages and trim_messages[0].get("role") == "tool":
                trim_messages.pop(0)

            response = self.openai_api_client.with_options(timeout=20.0).chat.completions.create(
                messages=trim_messages,
                **tools_para,
                **context.parameters,
            )
        except Exception as e:
            notification_signal.send(
                self,
                _async_wrapper=async_wrapper_without_loop,
                message={
                    "content":{
                        "content":{"error":e, "context":context},
                        "description":"An error occurred in communication with OpenAI initiated by user."
                    },
                    "flag":"info"
                }
            )
            OpenaiErrorHandler().openai_error_handle(error=e, context=context, event_loop=False)
            return

        response_stream_format = response_to_stream_format(mode=response_mode, response=response)
        
        callback = Callback(
            at_job_start=[
                {
                    "function": notification_signal.send,
                    "params": {
                        "args": (self,),
                        "kwargs": {
                            "_async_wrapper": async_wrapper_with_loop,
                            "message":{
                                "content":{
                                    "content":{"status": True, "context": context},
                                    "description":"Job status changed",
                                },
                                "flag":"info",
                            },
                        },
                    },
                },
            ],
            at_job_end=[
                {
                    "function": notification_signal.send_async,
                    "params": {
                        "args": (self,),
                        "kwargs": {
                            "_sync_wrapper": sync_wrapper,
                            "message":{
                                "content":{
                                    "content":{"status": False, "context": context},
                                    "description":"Job status changed",
                                },
                                "flag":"info",
                            },
                        },
                    },
                },
            ],
        )

        at_receiving_start = [
            {
                "function": notification_signal.send,
                "params": {
                    "args": (self,),
                    "kwargs": {
                        "_async_wrapper": async_wrapper_with_loop,
                        "message":{
                            "content":{
                                "content":{"context": context},
                                "description":"Starting to receive the original response message to the user",
                            },
                            "flag":"info",
                        },
                    },
                },
            },
        ]

        job = ResponseJob(
            manager=self.manager,
            response=response_stream_format,
            context=context,
            callback=callback,
            at_receiving_start=at_receiving_start,
        )
        self.manager.gk_kernel.commander.put_job_threadsafe(job)
    
    def chat_stream(self, context: OpenaiContext, message: dict | list[dict]) -> None:
        "stream version of chat function with openai"
        if isinstance(message, dict):
            messages_list = [message]
        else:
            messages_list = message

        if context.chat_context_saver_for_sending == "outer":
            chat_context_extend_for_sending_signal.send(
                self,
                _async_wrapper=async_wrapper_without_loop,
                message={
                    "content":{
                        "messages":messages_list,
                        "context":context,
                    },
                    "flag":"",
                }
            )
        else:
            self.chat_message_append(context=context, message=message)

        notification_signal.send(
            self,
            _async_wrapper=async_wrapper_without_loop,
            message={
                "content":{
                    "content":{"context":context},
                    "description":"Starting to send the original chat message from the user.",
                },
                "flag":"info",
            }
        )
        try:
            tools_para = {}
            if available_functions := self.manager.available_functions_meta:
                tools_para = {"tools": available_functions, "tool_choice": "auto"}
            
            offset_tokens_num = -tokens_num_for_functions_call(tools_para["tools"], model=context.parameters["model"])
            trim_messages = trim_excess_tokens(context, offset=offset_tokens_num)

            # Delete the tool reply messages at the beginning of the information list.
            # This is because if the information starts with a function reply message,
            # it indicates that the function call information has already been truncated.
            # The OpenAI API requires that function reply messages must be responses to function calls.
            # Therefore, if the function reply messages are not removed, it will result in an OpenAI API error.
            while trim_messages and trim_messages[0].get("role") == "tool":
                trim_messages.pop(0)

            response = self.openai_api_client.with_options(timeout=20.0).chat.completions.create(
                messages = trim_messages,
                **tools_para,
                **context.parameters,
                )
        except Exception as e:
            notification_signal.send(
                self,
                _async_wrapper=async_wrapper_without_loop,
                message={
                    "content":{
                        "content":{"error":e, "context":context},
                        "description":"An error occurred in communication with OpenAI initiated by user."
                    },
                    "flag":"info"
                }
            )
            OpenaiErrorHandler().openai_error_handle(error=e, context=context, event_loop=False)
            return
        
        callback = Callback(
            at_job_start=[
                {
                    "function": notification_signal.send,
                    "params": {
                        "args": (self,),
                        "kwargs": {
                            "_async_wrapper": async_wrapper_with_loop,
                            "message":{
                                "content":{
                                    "content":{"status": True, "context": context},
                                    "description":"Job status changed",
                                },
                                "flag":"info",
                            },
                        },
                    },
                },
            ],
            at_job_end=[
                {
                    "function": notification_signal.send_async,
                    "params": {
                        "args": (self,),
                        "kwargs": {
                            "_sync_wrapper": sync_wrapper,
                            "message":{
                                "content":{
                                    "content":{"status": False, "context": context},
                                    "description":"Job status changed",
                                },
                                "flag":"info",
                            },
                        },
                    },
                },
            ],
        )
        
        at_receiving_start = [
            {
                "function": notification_signal.send,
                "params": {
                    "args": (self,),
                    "kwargs": {
                        "_async_wrapper": async_wrapper_with_loop,
                        "message":{
                            "content":{
                                "content":{"context": context},
                                "description":"Starting to receive the original response message to the user",
                            },
                            "flag":"info",
                        },
                    },
                },
            },
        ]

        job = ResponseJob(
            manager=self.manager,
            response=response,
            context=context,
            callback=callback,
            at_receiving_start=at_receiving_start,
        )
        self.manager.gk_kernel.commander.put_job_threadsafe(job)


def response_to_stream_format(mode: Literal["no_function_call", "function_call"], response) -> list:
    if mode == "no_function_call":
        reply_content = response.choices[0].message.content
        finish_reason = response.choices[0].finish_reason
        result = [{"choices":[{"delta":{"content":""}, "finish_reason":"null"}]}, 
                                     {"choices":[{"delta":{"content":reply_content}, "finish_reason":"null"}]},
                                     {"choices":[{"delta":{}, "finish_reason":finish_reason}]}]
    elif mode == "function_call":
        reply_message = response.choices[0].message
        reply_content = reply_message.get("content")
        finish_reason = response.choices[0].finish_reason
        reply_function_call = reply_message.get("function_call")
        if reply_function_call is not None:
            name = reply_function_call.get("name")
            arguments = reply_function_call.get("arguments")
            result = [{"choices":[{"delta":{"content":None, "function_call":{"name":name}}, "finish_reason":"null"}]}, 
                                         {"choices":[{"delta":{"function_call": {"arguments":arguments}}, "finish_reason":"null"}]},
                                         {"choices":[{"delta":{}, "finish_reason":finish_reason}]}]
        else:
            result = [{"choices":[{"delta":{"content":""}, "finish_reason":"null"}]}, 
                                         {"choices":[{"delta":{"content":reply_content}, "finish_reason":"null"}]},
                                         {"choices":[{"delta":{}, "finish_reason":finish_reason}]}]
    return result


class OpenAIGroupTalk:
    def __init__(self, manager: ManagerInterface):
        self.manager = manager
        self.openai_api_client = openai_api_client(manager.dot_env_config_path)

    def talk_stream(self, group_talk_manager: GroupTalkManager, message_content: str) -> None:
        if group_talk_manager.state != "ACTIVE":
            group_talk_manager.speaking = None
            group_talk_manager.running = True
            group_talk_manager.user_talk_buffer.append(message_content)
            self.manager.gk_kernel.commander.put_job_threadsafe(group_talk_manager)
        else:
            group_talk_manager.user_talk_buffer.append(message_content)



import setuptools


setuptools.setup(
    packages=setuptools.find_packages(where="src"),
    package_dir={"": "src"},
    package_data={
        "gptui": [
            ".default_config.yml",
            "config.yml",
            "help.md",
            "**/*.txt",
            "**/*.tcss",
        ],
    },
)



import logging
import os

from textual import events
from textual.app import ComposeResult
from textual.containers import Grid
from textual.screen import ModalScreen
from textual.widgets import Button, Label, Input, MarkdownViewer

from .custom_tree import MyDirectoryTree
from ..utils.my_text import MyText as Text


gptui_logger = logging.getLogger("gptui_logger")


class CheckDialog(ModalScreen[bool]):
    """Check an action."""

    CSS = """
    CheckDialog {
        align: center middle;
    }
    CheckDialog > #dialog {
        grid-size: 2;
        grid-gutter: 1 2;
        grid-rows: 1fr 3;
        padding: 0 1;
        width: 60;
        height: 11;
        border: thick $background 80%;
        background: $surface;
    }
    CheckDialog #prompt {
        column-span: 2;
        height: 1fr;
        width: 1fr;
        content-align: center middle;
    }
    CheckDialog Button {
        width: 100%;
    }
    """

    def __init__(self, prompt: str | Text = "Are you sure?", yes_label: str = "Yse", no_label: str = "No", *args, **kwargs):
        self.prompt = prompt
        self.yes_label = yes_label
        self.no_label = no_label
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(self.prompt, id="prompt"),
            Button(self.yes_label, id="yes"),
            Button(self.no_label, id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            self.dismiss(True)
        elif event.button.id == "no":
            self.dismiss(False)


class InputDialog(ModalScreen[tuple[bool, str]]):
    """Ask whether to close the conversation."""

    CSS = """
    InputDialog {
        align: center middle;
    }
    InputDialog > #dialog {
        grid-size: 2;
        grid-gutter: 1 2;
        grid-rows: 1fr 2fr 3;
        padding: 0 1;
        width: 60;
        height: 11;
        border: thick $background 80%;
        background: $surface;
    }
    InputDialog #input {
        column-span: 2;
        height: 1fr;
        width: 1fr;
    }
    InputDialog #prompt {
        column-span: 2;
        height: 1fr;
        width: 1fr;
        content-align: center middle;
    }
    InputDialog Button{
        width: 100%;
    }
    """

    def __init__(self, prompt: str | Text = "Input:", placeholder: str = "", default_input: str | None = None, ok_label: str = "OK", cancel_label: str = "Cancel", *args, **kwargs):
        self.prompt = prompt
        self.placeholder = placeholder
        self.default_input = default_input
        self.ok_label = ok_label
        self.cancel_label = cancel_label
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        self.input = Input(placeholder=self.placeholder, id="input")
        yield Grid(
            Label(self.prompt, id="prompt"),
            self.input,
            Button(self.ok_label, id="ok"),
            Button(self.cancel_label, id="cancel"),
            id="dialog",
        )
    
    def on_mount(self):
        if self.default_input:
            self.input.value = self.default_input

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "ok":
            self.dismiss((True, self.input.value))
        elif event.button.id == "cancel":
            self.dismiss((False, self.input.value))


class SelectPathDialog(ModalScreen[tuple[bool, str]]):
    """Ask whether to close the conversation."""

    CSS = """
    SelectPathDialog {
        align: center middle;
    }
    SelectPathDialog > #dialog {
        grid-size: 2;
        grid-gutter: 1 2;
        grid-rows: 1 1fr 2fr 3;
        padding: 0 1;
        width: 60;
        height: 20;
        border: thick $background 80%;
        background: $surface;
    }
    SelectPathDialog #input {
        column-span: 2;
        height: 1;
        width: 1fr;
    }
    SelectPathDialog #select_path_directory_tree {
        column-span: 2;
        height: 1fr;
        width: 1fr;
    }
    SelectPathDialog #prompt {
        column-span: 2;
        height: 1;
        width: 1fr;
        content-align: center middle;
    }
    SelectPathDialog Button{
        width: 100%;
    }
    """

    def __init__(
        self,
        root_directory_path: str,
        prompt: str | Text = "Input:",
        placeholder: str = "",
        ok_label: str = "OK",
        cancel_label: str = "Cancel",
        *args,
        **kwargs
        ):
        self.root_directory_path = root_directory_path
        self.prompt = prompt
        self.placeholder = placeholder
        self.ok_label = ok_label
        self.cancel_label = cancel_label
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        self.input = Input(placeholder=self.placeholder, id="input")
        yield Grid(
            Label(self.prompt, id="prompt"),
            self.input,
            MyDirectoryTree(self.root_directory_path, self.root_directory_path, id="select_path_directory_tree"),
            Button(self.ok_label, id="ok"),
            Button(self.cancel_label, id="cancel"),
            id="dialog",
        )

    def on_tree_node_highlighted(self, event) -> None:
        selected_path = event.node.data.path
        if selected_path.is_file():
            self.input.value = str(selected_path.parent) + os.sep
        if selected_path.is_dir():
            self.input.value = str(selected_path) + os.sep
    
    async def on_input_submitted(self, message) -> None:
        self.dismiss((True, message.value))

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "ok":
            self.dismiss((True, self.input.value))
        elif event.button.id == "cancel":
            self.dismiss((False, self.input.value))


class MarkdownPreview(ModalScreen):
    
    CSS = """
    MarkdownPreview {
        align: center middle;
    }
    MarkdownPreview MarkdownViewer {
        width: 50%;
        height: 50%;
        border: ascii $accent;
    }
    """

    BINDINGS = [
        ("space", "quit", "Quit"),
    ]

    def __init__(self, markdown: str | None = None, previewer_title: str | None = None, *args, **kwargs):
        self.markdown = markdown
        self.previewer_title = previewer_title
        super().__init__(*args, **kwargs)

    def on_mount(self) -> None:
        previewer = self.query_one(MarkdownViewer)
        if self.previewer_title is not None:
            previewer.border_title = self.previewer_title
        previewer.border_subtitle = "Press 'space' to close this previewer"
    
    def compose(self) -> ComposeResult:
        yield MarkdownViewer(self.markdown, show_table_of_contents=False)

    def action_quit(self) -> None:
        self.app.pop_screen()


class HotKey(ModalScreen[str]):

    CSS = """
    HotKey {
        align: center bottom;
    }
    HotKey Label {
        width: 80;
        height: 12;
        text-align: center;
        border: ascii $accent;
    }
    """

    BINDINGS = [
        ("escape,ctrl+underscore", "quit", "Quit"),
    ]

    def __init__(self, display_content: str | Text, *args, **kwargs):
        self.display_content = display_content
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        yield Label(self.display_content)

    def action_quit(self) -> None:
        self.app.pop_screen()

    def on_key(self, event: events.Key) -> None:
        self.dismiss(event.key)



from rich.console import Console

from .my_text import MyText


def my_line_count(content: MyText, width: int, console=Console()) -> int:
    lines = content.split(allow_blank=True)
    num_count = 0
    for line in lines:
        num_count += len(line.wrap(console, width))
    return num_count



import json
import logging

from semantic_kernel.skill_definition import sk_function

from ....gptui_kernel import Kernel
from ....gptui_kernel.manager import auto_init_params


gptui_logger = logging.getLogger("gptui_logger")


class ConversationService:
    def __init__(self, manager):
        self.gk_kernel = Kernel(manager.dot_env_config_path)

    @auto_init_params("0")
    @classmethod
    def get_init_params(cls, manager) -> tuple:
        return (manager,)

    @sk_function(
        description="Generate a title for given conversation context. The conversation context is a json string converted from a conversation dict with openai gpt.",
        name="conversation_title",
        input_description="The json string of conversation which need a title.",
    )
    async def conversation_title(self, chat_context_json_str: str) -> str:
        def chat_context_to_string(chat_context_json_str: str) -> str:
            chat_context_json = json.loads(chat_context_json_str)
            chat_context = ''
            assert isinstance(chat_context_json, list)
            for piece in chat_context_json:
                chat_context += piece["role"] + ": " + str(piece["content"] or piece.get("tool_calls") or "") + "\n\n"
            return chat_context[:1000]
        
        sk_prompt = (
            "Generate a concise and clear title for the following chat record. "
            "The title should be as brief as possible, not exceeding ten English words or twenty Chinese characters, " 
            "and the language of the title should be consistent with the content of the chat. "
            "Do not have line breaks '\\n'. "
            "chat record: {{$INPUT}}\n"
            "title:"
        )
        
        chat_context = chat_context_to_string(chat_context_json_str)
        make_title_function = self.gk_kernel.sk_kernel.create_semantic_function(sk_prompt, max_tokens=50)
        name = await make_title_function.invoke_async(chat_context)
        name = str(name)
        if name.startswith('"'):
            name = name[1:]
            if name.endswith('"'):
                name = name[:-1]
        return name



import os
import sys
import inspect
import logging
from abc import ABCMeta, abstractmethod
from importlib import import_module
from typing import Type

from agere.commander import Job

from .kernel import KernelInterface, Kernel, PluginMeta, TemporarySysPath
from .manager_exceptions import HandlerNotRegisterError, JobNotRegisterError
from .null_logger import get_null_logger


class ManagerInterface(metaclass=ABCMeta):
    @property
    @abstractmethod
    def gk_kernel(self) -> KernelInterface:
        ...

    @gk_kernel.setter
    @abstractmethod
    def gk_kernel(self, kernel: KernelInterface) -> None:
        ...

    @property
    @abstractmethod
    def services(self) -> KernelInterface:
        ...
    
    @services.setter
    @abstractmethod
    def services(self, kernel: KernelInterface) -> None:
        ...
    
    @property
    @abstractmethod
    def client(self) -> Type:
        ...

    @client.setter
    @abstractmethod
    def client(self, client) -> None:
        ...

    @property
    @abstractmethod
    def jobs(self) -> dict:
        ...

    @property
    @abstractmethod
    def handlers(self) -> dict:
        ...

    @staticmethod
    @abstractmethod
    def make_gk_kernel(dot_env_config_path: str | None = None, logger: logging.Logger | None = None) -> KernelInterface:
        ...

    @property
    @abstractmethod
    def available_functions_meta(self) -> list[dict]:
        ...

    @property
    @abstractmethod
    def available_functions_link(self) -> list[dict]:
        ...

    @abstractmethod
    def load_services(self, where, skill_name) -> None:
        ...

    @abstractmethod
    def register_jobs(self, module_name) -> None:
        ...

    @abstractmethod
    def register_handlers(self, module_name) -> None:
        ...

    @abstractmethod
    def get_job(self, job: str) -> Type:
        ...

    @abstractmethod
    def get_handler(self, handler: str) -> Type:
        ...

    @abstractmethod
    def add_plugins(self, plugins: list[tuple] | tuple) -> None:
        ...

    @abstractmethod
    def remove_plugins(self, plugins: list[tuple] | tuple) -> None:
        ...

    @abstractmethod
    def overwrite_plugins(self, plugins: list[tuple] | tuple) -> None:
        ...

    @abstractmethod
    def scan_plugins(self, path) -> tuple[list, list]:
        ...

    @property
    @abstractmethod
    def dot_env_config_path(self) -> str | None:
        ...


class Manager(ManagerInterface):
    def __init__(
        self,
        client,
        kernel: KernelInterface | None = None,
        service_kernel: KernelInterface | None = None,
        dot_env_config_path: str | None = None,
        logger: logging.Logger | None = None
    ):
        self._dot_env_config_path = dot_env_config_path
        self.logger = logger or get_null_logger()
        self._gk_kernel = kernel or Kernel(dot_env_config_path=dot_env_config_path, logger=self.logger)
        self._services = service_kernel or Kernel(dot_env_config_path=dot_env_config_path, logger=self.logger)
        self._client = client
        self._jobs = {}
        self._handlers = {}

    @property
    def dot_env_config_path(self) -> str | None:
        return self._dot_env_config_path

    @property
    def gk_kernel(self) -> KernelInterface:
        return self._gk_kernel
    
    @gk_kernel.setter
    def gk_kernel(self, kernel: KernelInterface) -> None:
        self._gk_kernel = kernel

    @property
    def services(self) -> KernelInterface:
        return self._services

    @services.setter
    def services(self, service_kernel: KernelInterface) -> None:
        self._services = service_kernel

    @property
    def client(self):
        return self._client

    @client.setter
    def client(self, client_) -> None:
        self.client = client_

    @property
    def jobs(self) -> dict:
        return self._jobs

    @property
    def handlers(self) -> dict:
        return self._handlers

    @staticmethod
    def make_gk_kernel(dot_env_config_path: str | None = None, logger: logging.Logger | None = None) -> KernelInterface:
        return Kernel(dot_env_config_path=dot_env_config_path, logger=logger)
    
    @property
    def available_functions_meta(self) -> list[dict]:
        meta, _ = self.gk_kernel.llm_function_call.sk_plugins_to_openai_functions()
        return meta

    @property
    def available_functions_link(self) -> dict:
        _, link = self.gk_kernel.llm_function_call.sk_plugins_to_openai_functions()
        return link

    def load_services(self, where, skill_name) -> None:
        if isinstance(where, str):
            self.services.sk_kernel.import_semantic_skill_from_directory(where, skill_name)
        else:
            self.services.sk_kernel.import_skill(where, skill_name=skill_name)

    def register_jobs(self, module_name) -> None:
        module = import_module(module_name)
        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj) or inspect.isfunction(obj):
                self._jobs[name] = obj

    def register_handlers(self, module_name) -> None:
        module = import_module(module_name)
        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj) or inspect.isfunction(obj):
                self._handlers[name] = obj

    def get_handler(self, handler: str):
        try:
            aim_handler = self.handlers[handler]
        except KeyError:
            raise HandlerNotRegisterError(handler=handler, manager=self)
        return aim_handler

    def get_job(self, job: str) -> Job:
        try:
            aim_job = self.jobs[job]
        except KeyError:
            raise JobNotRegisterError(job=job, manager=self)
        return aim_job
    
    def add_plugins(self, plugins: list[tuple] | tuple) -> None:
        if isinstance(plugins, tuple):
            plugins_list = [plugins]
        else:
            plugins_list = plugins
        self.gk_kernel.add_plugins(plugins_list)
    
    def remove_plugins(self, plugins: list[tuple] | tuple) -> None:
        if isinstance(plugins, tuple):
            plugins_list = [plugins]
        else:
            plugins_list = plugins
        self.gk_kernel.remove_plugins(plugins_list)

    def overwrite_plugins(self, plugins: list[tuple] | tuple) -> None:
        if isinstance(plugins, tuple):
            plugins_list = [plugins]
        else:
            plugins_list = plugins
        self.gk_kernel.overwrite_plugins(plugins_list)

    def scan_plugins(self, path) -> tuple[list, list]:
        children = os.listdir(path)
        dirs = [d for d in children if os.path.isdir(os.path.join(path, d))]
        files = [os.path.splitext(f)[0] for f in children if os.path.isfile(os.path.join(path, f))]
        
        def check_semantic_plugin(dir_path: str) -> bool:
            children = os.listdir(dir_path)
            dirs = [d for d in children if os.path.isdir(os.path.join(dir_path, d))]
            for dir in dirs:
                contents = os.listdir(os.path.join(dir_path, dir))
                files = [f for f in contents if os.path.isfile(os.path.join(dir_path, dir, f))]
                if {"skprompt.txt", "config.json"}.issubset(set(files)):
                    return True
            return False

        semantic_plugins_info_list = [(path, plugin_name) for plugin_name in dirs if check_semantic_plugin(os.path.join(path, plugin_name))]
        semantic_plugins_list = [PluginMeta(plugin_info) for plugin_info in semantic_plugins_info_list]

        with TemporarySysPath():
            if path not in sys.path:
                sys.path.append(path)
            native_plugins_list = []
            for file in files:
                module = import_module(file)
                for name, obj in inspect.getmembers(module):
                    if inspect.isclass(obj) and obj.__module__ == module.__name__:
                        params = self._check_auto_init_params(obj)
                        if params is None:
                            native_plugins_list.append(PluginMeta((path, file, name)))
                        else:
                            native_plugins_list.append(PluginMeta((path, file, name, params)))
        
        return semantic_plugins_list, native_plugins_list

    def _check_auto_init_params(self, plugin_class) -> tuple | None:
        params = None
        for _, method in plugin_class.__dict__.items():
            if hasattr(method, "_auto_init_params_") and method._auto_init_params_ is True:
                mode = method._auto_init_params_mode_
                if not isinstance(method, classmethod):
                    raise TypeError(f"{method} is not a classmethod.")
                if mode == "0":
                    params = method.__func__(plugin_class, self)
                else:
                    raise ValueError(f"Get error _auto_init_params_mode_: {mode}.")
        return params

def auto_init_params(mode: str = "0"):
    """
    Decorator:
        Mark the method for automatically obtaining initialization parameters,
        and set the mode of the callback parameter for this method.
        Currently, only mode "0" is supported, which requires the "manager" to return itself.
        It should be used outer of @classmethod.
    Usage:
        @auto_init_params("0")
        @classmethod
        def get_init_params(cls, manager) -> tuple:
            ...
    """
    def decorator(func):
        func._auto_init_params_ = True
        func._auto_init_params_mode_ = mode
        return func
    return decorator



import logging
import os

from .drivers import CopyCode, TextToSpeak, VoiceRecordStart, VoiceRecordQuit
from .driver_interface import DriverInterface


gptui_logger = logging.getLogger("gptui_logger")


class DriverManager:
    def __init__(self, app):
        self.app = app
        self.terminal = app.config.get("terminal")
        self.os = app.config.get("os")
        self._register_driver_method()

    def register_driver(self, driver_method_name: str, driver: DriverInterface) -> None:
        if hasattr(self, driver_method_name):
            gptui_logger.warning("A driver method with the same name already exists; it will be overwritten.")
        setattr(self, driver_method_name, driver)
    
    def _register_driver_method(self):
        self.copy_code = CopyCode(self.os)
        self.tts = TextToSpeak(
            platform=self.os,
            dot_env_path=self.app.config["dot_env_path"],
            temp_dir=os.path.join(self.app.config["workpath"], "temp"),
        )
        self.voice_record_start = VoiceRecordStart(self.os)
        self.voice_record_quit = VoiceRecordQuit(self.os)



import json
import logging

from semantic_kernel.orchestration.sk_context import SKContext
from semantic_kernel.skill_definition import sk_function, sk_function_context_parameter

from gptui.gptui_kernel.manager import auto_init_params


gptui_logger = logging.getLogger("gptui_logger")


class MemoryRecall:
    def __init__(self, manager):
        self.manager = manager
    
    @auto_init_params("0")
    @classmethod
    def get_init_params(cls, manager) -> tuple:
        return (manager,)

    @sk_function(
        description="Recall the specified content from the memory store.",
        name="recall_memory",
    )
    @sk_function_context_parameter(
        name="query",
        description="Topics, questions, etc., that one needs to recall."
    )
    @sk_function_context_parameter(
        name="max_recallable_entries",
        description="Maximum number of recallable information entries.",
        default_value="1"
    )
    @sk_function_context_parameter(
        name="openai_context",
        description=(
            "The dictionary string version of the OpenaiContext instance. "
            "This is a special parameter that typically doesn't require manual intervention, as it is usually automatically managed."
            "Unless there's a clear intention, please keep its default value."
        ),
        default_value="AUTO"
    )
    async def recall_memory(self, context: SKContext) -> str:
        query = context["query"]
        max_recallable_entries = int(context["max_recallable_entries"])
        openai_context_dict = json.loads(str(context["openai_context"]))
        conversation_id = openai_context_dict["id"]
        semantic_memory = self.manager.services.sk_kernel.memory
        try:
            result = await semantic_memory.search_async(str(conversation_id), query, limit=max_recallable_entries, min_relevance_score=0.7)
        except Exception as e:
            gptui_logger.error(f"Error occurred when recall memory. Error: {e}")
            return "An error occurred during the query, please try again later."
        result_str = ""
        for memory in result:
            result_str += memory.id + "\n"
        if not result_str:
            result_str = "No relevant information was found"
        gptui_logger.info(f"Recall memory result:\nconversation_id: {conversation_id}\nResult: {result_str}")
        return result_str



import asyncio
import logging
from abc import ABC, abstractmethod

from rich.text import TextType


gptui_logger = logging.getLogger("gptui_logger")


class Happy(ABC):

    @abstractmethod
    def refresh(self, content, apple_width: int, apple_height: int) -> None:
        ...

    @property
    @abstractmethod
    def canvas_width(self) -> int:
        ...
    
    @property
    @abstractmethod
    def canvas_height(self) -> int:
        ...


class Apple(ABC):

    @abstractmethod
    def frame(self, inp) -> tuple[bool, tuple[float, TextType]]:
        ...

    @property
    @abstractmethod
    def canvas_width(self) -> int:
        ...
    
    @property
    @abstractmethod
    def canvas_height(self) -> int:
        ...


class Horse:
    def __init__(self):
        self.input = None
        self.happy = None
        self.apple = None
        self.run_status = False
        self._stop_flag = False
        self._stop_async_flag = False
        self._stop_async_event = asyncio.Event()

    def set_happy(self, happy: Happy) -> None:
        self.happy = happy

    def refresh_input(self, user_input: str) -> None:
        self.input = user_input
    
    def stop(self):
        self._stop_flag = True

    async def stop_async(self):
        if self.run_status is False:
            return
        self._stop_async_event.clear()
        self._stop_async_flag = True
        await self._stop_async_event.wait()

    async def run(self, apple: Apple, size_check: bool = True) -> bool:
        self.run_status = True
        assert self.happy is not None
        self._stop_flag = False
        self._stop_async_flag = False
        self.apple = apple
        if size_check:
            if (apple.canvas_width > self.happy.canvas_width) or (apple.canvas_height > self.happy.canvas_height):
                self.run_status = False
                return False
        
        status = True
        while status:
            inp = self.input
            self.input = None
            status, frame_info = apple.frame(inp)
            if self._stop_flag or self._stop_async_flag:
                break
            await self.frame_handle(frame_info)
        
        if self._stop_async_flag:
            self._stop_async_event.set()
        self.run_status = False
        return True

    async def frame_handle(self, frame_info: tuple[float, TextType] | None) -> None:
        if frame_info is None:
            return
        interval, frame = frame_info
        await asyncio.sleep(interval)
        self.happy_render(frame)

    def happy_render(self, content: TextType):
        assert self.happy is not None
        assert self.apple is not None
        self.happy.refresh(content, self.apple.canvas_width, self.apple.canvas_height)



import logging

from openai.types.chat import ChatCompletionMessageParam

from .context import OpenaiContext
from .utils.tokens_num import tokens_num_from_string, tokens_num_from_chat_context


gptui_logger = logging.getLogger("gptui_logger")


def find_position(lst, num):
    """
    Finds the farthest left position in the list where the sum of all elements after that position is less than the given number 'num'.
    If you move one position backward, the sum of all elements after that new position will be greater than 'num'.
    
    Args:
        - lst {list}: The list of integers.
        - num {int}: The target number.
    
    Returns:
        int: The 1-based index of the position found, (index + 1)
        return 0 if the sum of all integers is less than num,
        return the length of lst if the last integers is greater than num.
    """
    prefix_sum = [0]
    for x in lst:
        prefix_sum.append(prefix_sum[-1] + x)

    left, right = 0, len(prefix_sum) - 1

    while left < right:
        mid = (left + right) // 2
        if prefix_sum[-1] - prefix_sum[mid] < num:
            right = mid
        else:
            left = mid + 1

    return left

def trim_excess_tokens(
    context: OpenaiContext,
    max_tokens_num: int | None = None,
    offset: int = 0
) -> list[ChatCompletionMessageParam]:
    """Truncate the given context according to max_tokens_num, only retaining the last part.

    It will return a new chat_context list and not change the original chat_context.

    Args:
        - context (OpenaiContext): the context need to be trimmed.
        - max_tokens_num (int): the max tokens number allowed.
        - offset (int): for positive value, increase max_tokens_num; for negative value, decrease max_tokens_num.

    Retruns:
        list[dict]: truncated chat_context of context.
    """
    tokens_num_list = context.tokens_num_list
    if max_tokens_num is None:
        max_tokens_num = context.max_sending_tokens_num
    assert max_tokens_num is not None
    assert context.chat_context is not None
    num_after_offset = max_tokens_num + offset
    if num_after_offset <= 0:
        gptui_logger.warning(
            "The valid token length is less than zero, only the last message is kept. "
            "This could likely lead to a token lenght exceeding the limit error."
        )
        return context.chat_context[-1:]
    position = find_position(lst=tokens_num_list, num=num_after_offset)
    if position >= len(tokens_num_list):
        model = context.parameters["model"]
        trim_status = True
        new_tokens_num = num_after_offset
        out_dict = context.chat_context[-1:][0] # Don't change the original context.
        while trim_status:
            new_tokens_num -= 5
            if new_tokens_num <= 0: # The number 5 is the assumed additional tokens count in message dict compared to message content
                out_dict["content"] = ""
                return [out_dict]
            out_dict_content = out_dict["content"]
            assert isinstance(out_dict_content, str)
            trim_string = trim_string_by_tokens(out_dict_content, max_tokens=new_tokens_num, model=model)
            out_dict["content"] = trim_string
            if tokens_num_from_chat_context([out_dict], model=model) < num_after_offset:
                trim_status = False
        return [out_dict]
    return context.chat_context[position:]

def trim_string_by_tokens(string: str, max_tokens: int, model: str) -> str:
    """trims the input string based on a specified maximum token count.
        - If the overall token count of the input string is less than or equal to the specified maximum token count, the function returns the original string as is.
        - If the token count of the input string exceeds the specified maximum, the function trims words from the beginning of the string progressively until the token count does not exceed the limit.

    Parameters:
        - string (str): The input string to be trimmed.
        - max_tokens (int): The allowable maximum token count.

    Returns:
        - str: The trimmed string where the token count does not surpass the specified maximum token count.
    """
    words = string.split()
    if tokens_num_from_string(string, model) <= max_tokens:
        return string

    left, right = 0, len(words)
    
    while left < right:
        mid = (left + right) // 2
        current_str = ' '.join(words[mid:])
        current_tokens = tokens_num_from_string(current_str, model)
        
        if current_tokens == max_tokens:
            return current_str
        elif current_tokens < max_tokens:
            right = mid
        else:
            left = mid + 1

    if right == len(words):
        return ""
    return ' '.join(words[right:])
